# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""An executable class to run agents in the simulator."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import time

from absl import flags
from dopamine.discrete_domains import checkpointer
import gin.tf
from gym import spaces
import numpy as np
import tensorflow as tf


flags.DEFINE_bool(
    'debug_mode', False,
    'If set to true, the agent will output in-episode statistics '
    'to Tensorboard. Disabled by default as this results in '
    'slower training.')
flags.DEFINE_string('agent_name', None, 'Name of the agent.')
flags.DEFINE_string('base_dir', None,
                    'Base directory to host all required sub-directories.')
flags.DEFINE_string(
    'environment_name', 'interest_evolution',
    'The environment with which to run the experiment. Supported choices are '
    '{interest_evolution, interest_exploration}.')
flags.DEFINE_string(
    'episode_log_file', '',
    'Filename under base_dir to output simulated episodes in SequenceExample.')
flags.DEFINE_multi_string(
    'gin_files', [], 'List of paths to gin configuration files (e.g.'
    '"third_party/py/dopamine/agents/dqn/dqn.gin").')
flags.DEFINE_multi_string(
    'gin_bindings', [],
    'Gin bindings to override the values set in the config files '
    '(e.g. "runner_lib.Runner.max_steps_per_episode=100')


FLAGS = flags.FLAGS


def load_gin_configs(gin_files, gin_bindings):
  """Loads gin configuration files.

  Args:
    gin_files: list, of paths to the gin configuration files for this
      experiment.
    gin_bindings: list, of gin parameter bindings to override the values in the
      config files.
  """
  gin.parse_config_files_and_bindings(
      gin_files, bindings=gin_bindings, skip_unknown=False)


@gin.configurable
class Runner(object):
  """Object that handles running experiments.

  Here we use the term 'experiment' to mean simulating interactions between the
  agent and the environment and reporting some statistics pertaining to these
  interactions.
  """

  _output_dir = None
  _checkpoint_dir = None
  _agent = None
  _checkpointer = None

  def __init__(self,
               base_dir,
               create_agent_fn,
               env,
               episode_log_file='',
               checkpoint_file_prefix='ckpt',
               max_steps_per_episode=27000):
    """Initializes the Runner object in charge of running a full experiment.

    Args:
      base_dir: str, the base directory to host all required sub-directories.
      create_agent_fn: A function that takes as args a Tensorflow session and an
        environment, and returns an agent.
      env: A Gym environment for running the experiments.
      episode_log_file: Path to output simulated episodes in tf.SequenceExample.
        Disable logging if episode_log_file is an empty string.
      checkpoint_file_prefix: str, the prefix to use for checkpoint files.
      max_steps_per_episode: int, maximum number of steps after which an episode
        terminates.
    """
    tf.compat.v1.logging.info('max_steps_per_episode = %s',
                              max_steps_per_episode)

    if base_dir is None:
      raise ValueError('Missing base_dir.')

    self._base_dir = base_dir
    self._create_agent_fn = create_agent_fn
    self._env = env
    self._checkpoint_file_prefix = checkpoint_file_prefix
    self._max_steps_per_episode = max_steps_per_episode
    self._episode_log_file = episode_log_file
    self._episode_writer = None

  def _set_up(self, eval_mode):
    """Sets up the runner by creating and initializing the agent."""
    # Reset the tf default graph to avoid name collisions from previous runs
    # before doing anything else.
    tf.compat.v1.reset_default_graph()
    self._summary_writer = tf.compat.v1.summary.FileWriter(self._output_dir)
    if self._episode_log_file:
      self._episode_writer = tf.io.TFRecordWriter(
          os.path.join(self._output_dir, self._episode_log_file))
    # Set up a session and initialize variables.
    self._sess = tf.compat.v1.Session(
        config=tf.compat.v1.ConfigProto(allow_soft_placement=True))
    self._agent = self._create_agent_fn(
        self._sess,
        self._env,
        summary_writer=self._summary_writer,
        eval_mode=eval_mode)
    self._summary_writer.add_graph(graph=tf.compat.v1.get_default_graph())
    self._sess.run(tf.compat.v1.global_variables_initializer())
    self._sess.run(tf.compat.v1.local_variables_initializer())

  def _initialize_checkpointer_and_maybe_resume(self, checkpoint_file_prefix):
    """Reloads the latest checkpoint if it exists.

    This method will first create a `Checkpointer` object and then call
    `checkpointer.get_latest_checkpoint_number` to determine if there is a valid
    checkpoint in self._checkpoint_dir, and what the largest file number is.
    If a valid checkpoint file is found, it will load the bundled data from this
    file and will pass it to the agent for it to reload its data.
    If the agent is able to successfully unbundle, this method will increase and
    return the iteration number keyed by 'current_iteration' and the step number
    keyed by 'total_steps' as the return values.

    Args:
      checkpoint_file_prefix: str, the checkpoint file prefix.
    Returns:
      start_iteration: The iteration number to be continued after the latest
        checkpoint.
      start_step: The step number to be continued after the latest checkpoint.
    """
    self._checkpointer = checkpointer.Checkpointer(self._checkpoint_dir,
                                                   checkpoint_file_prefix)
    start_iteration = 0
    start_step = 0
    # Check if checkpoint exists.
    # Note that the existence of checkpoint 0 means that we have finished
    # iteration 0 (so we will start from iteration 1).
    latest_checkpoint_version = checkpointer.get_latest_checkpoint_number(
        self._checkpoint_dir)
    if latest_checkpoint_version >= 0:
      assert not self._episode_writer, 'Can only log episodes from scratch.'
      experiment_data = self._checkpointer.load_checkpoint(
          latest_checkpoint_version)
      start_iteration = experiment_data['current_iteration'] + 1
      del experiment_data['current_iteration']
      start_step = experiment_data['total_steps'] + 1
      del experiment_data['total_steps']
      if self._agent.unbundle(self._checkpoint_dir, latest_checkpoint_version,
                              experiment_data):
        tf.compat.v1.logging.info(
            'Reloaded checkpoint and will start from '
            'iteration %d', start_iteration)
    return start_iteration, start_step

  def _log_one_step(self, user_obs, doc_obs, slate, responses, reward,
                    is_terminal, sequence_example):
    """Adds one step of agent-environment interaction into SequenceExample.

    Args:
      user_obs: An array of floats representing user state observations
      doc_obs: A list of observations of the documents
      slate: An array of indices to doc_obs
      responses: A list of observations of responses for items in the slate
      reward: A float for the reward returned after this step
      is_terminal: A boolean for whether a terminal state has been reached
      sequence_example: A SequenceExample proto for logging current episode
    """

    def _add_float_feature(feature, values):
      feature.feature.add(float_list=tf.train.FloatList(value=values))

    def _add_int64_feature(feature, values):
      feature.feature.add(int64_list=tf.train.Int64List(value=values))

    if self._episode_writer is None:
      return
    fl = sequence_example.feature_lists.feature_list
    _add_float_feature(
        fl['user'],
        spaces.flatten(self._env.observation_space.spaces['user'], user_obs))
    for i, doc in enumerate(list(doc_obs.values())):
      doc_space = list(
          self._env.observation_space.spaces['doc'].spaces.values())[i]
      _add_float_feature(fl['doc_%d' % i], spaces.flatten(doc_space, doc))
    _add_int64_feature(fl['slate'], slate)
    for i, response in enumerate(responses):
      resp_space = self._env.observation_space.spaces['response'][0]
      for k in response:
        _add_float_feature(fl['response_%d_%s' % (i, k)],
                           spaces.flatten(resp_space, response))
    _add_float_feature(fl['reward'], [reward])
    _add_int64_feature(fl['is_terminal'], [is_terminal])

  def _run_one_episode(self):
    """Executes a full trajectory of the agent interacting with the environment.

    Returns:
      The number of steps taken and the total reward.
    """
    step_number = 0
    total_reward = 0.

    start_time = time.time()

    sequence_example = tf.train.SequenceExample()
    observation = self._env.reset()
    action = self._agent.begin_episode(observation)

    # Keep interacting until we reach a terminal state.
    while True:
      last_observation = observation
      observation, reward, done, info = self._env.step(action)
      self._log_one_step(last_observation['user'], last_observation['doc'],
                         action, observation['response'], reward, done,
                         sequence_example)
      # Update environment-specific metrics with responses to the slate.
      self._env.update_metrics(observation['response'], info)

      total_reward += reward
      step_number += 1

      if done:
        break
      elif step_number == self._max_steps_per_episode:
        # Stop the run loop once we reach the true end of episode.
        break
      else:
        action = self._agent.step(reward, observation)

    self._agent.end_episode(reward, observation)
    if self._episode_writer is not None:
      self._episode_writer.write(sequence_example.SerializeToString())

    time_diff = time.time() - start_time
    self._update_episode_metrics(
        episode_length=step_number,
        episode_time=time_diff,
        episode_reward=total_reward)

    return step_number, total_reward

  def _initialize_metrics(self):
    """Initializes the metrics."""
    self._stats = {
        'episode_length': [],
        'episode_time': [],
        'episode_reward': [],
    }
    # Initialize environment-specific metrics.
    self._env.reset_metrics()

  def _update_episode_metrics(self, episode_length, episode_time,
                              episode_reward):
    """Updates the episode metrics with one episode."""

    self._stats['episode_length'].append(episode_length)
    self._stats['episode_time'].append(episode_time)
    self._stats['episode_reward'].append(episode_reward)

  def _write_metrics(self, step, suffix):
    """Writes the metrics to Tensorboard summaries."""

    def add_summary(tag, value):
      summary = tf.compat.v1.Summary(value=[
          tf.compat.v1.Summary.Value(
              tag=tag + '/' + suffix, simple_value=value)
      ])
      self._summary_writer.add_summary(summary, step)

    num_steps = np.sum(self._stats['episode_length'])
    time_per_step = np.sum(self._stats['episode_time']) / num_steps

    add_summary('TimePerStep', time_per_step)
    add_summary('AverageEpisodeLength', np.mean(self._stats['episode_length']))
    add_summary('AverageEpisodeRewards', np.mean(self._stats['episode_reward']))
    add_summary('StdEpisodeRewards', np.std(self._stats['episode_reward']))

    # Environment-specific Tensorboard summaries.
    self._env.write_metrics(add_summary)

    self._summary_writer.flush()

  def _checkpoint_experiment(self, iteration, total_steps):
    """Checkpoints experiment data.

    Args:
      iteration: int, iteration number for checkpointing.
      total_steps: int, total number of steps for all iterations so far.
    """
    experiment_data = self._agent.bundle_and_checkpoint(self._checkpoint_dir,
                                                        iteration)
    if experiment_data:
      experiment_data['current_iteration'] = iteration
      experiment_data['total_steps'] = total_steps
      self._checkpointer.save_checkpoint(iteration, experiment_data)


@gin.configurable
class TrainRunner(Runner):
  """Object that handles running the training.

  See main.py for a simple example to train an agent.
  """

  def __init__(self, max_training_steps=250000, num_iterations=100,
               checkpoint_frequency=1, **kwargs):
    tf.compat.v1.logging.info(
        'max_training_steps = %s, number_iterations = %s,'
        'checkpoint frequency = %s iterations.', max_training_steps,
        num_iterations, checkpoint_frequency)

    super(TrainRunner, self).__init__(**kwargs)
    self._max_training_steps = max_training_steps
    self._num_iterations = num_iterations
    self._checkpoint_frequency = checkpoint_frequency

    self._output_dir = os.path.join(self._base_dir, 'train')
    self._checkpoint_dir = os.path.join(self._output_dir, 'checkpoints')

    self._set_up(eval_mode=False)

  def run_experiment(self):
    """Runs a full experiment, spread over multiple iterations."""
    tf.compat.v1.logging.info('Beginning training...')
    start_iter, total_steps = self._initialize_checkpointer_and_maybe_resume(
        self._checkpoint_file_prefix)
    if self._num_iterations <= start_iter:
      tf.compat.v1.logging.warning('num_iterations (%d) < start_iteration(%d)',
                                   self._num_iterations, start_iter)
      return

    for iteration in range(start_iter, self._num_iterations):
      tf.compat.v1.logging.info('Starting iteration %d', iteration)
      total_steps = self._run_train_phase(total_steps)
      if iteration % self._checkpoint_frequency == 0:
        self._checkpoint_experiment(iteration, total_steps)

  def _run_train_phase(self, total_steps):
    """Runs training phase and updates total_steps."""

    self._initialize_metrics()

    num_steps = 0

    while num_steps < self._max_training_steps:
      episode_length, _ = self._run_one_episode()
      num_steps += episode_length

    total_steps += num_steps
    self._write_metrics(total_steps, suffix='train')
    return total_steps


@gin.configurable
class EvalRunner(Runner):
  """Object that handles running the evaluation.

  See main.py for a simple example to evaluate an agent.
  """

  def __init__(self,
               max_eval_episodes=125000,
               test_mode=False,
               min_interval_secs=30,
               train_base_dir=None,
               **kwargs):
    tf.compat.v1.logging.info('max_eval_episodes = %s', max_eval_episodes)
    super(EvalRunner, self).__init__(**kwargs)
    self._max_eval_episodes = max_eval_episodes
    self._test_mode = test_mode
    self._min_interval_secs = min_interval_secs

    self._output_dir = os.path.join(self._base_dir,
                                    'eval_%s' % max_eval_episodes)
    tf.io.gfile.makedirs(self._output_dir)
    if train_base_dir is None:
      train_base_dir = self._base_dir
    self._checkpoint_dir = os.path.join(train_base_dir, 'train', 'checkpoints')

    self._set_up(eval_mode=True)

  def run_experiment(self):
    """Runs a full experiment, spread over multiple iterations."""
    tf.compat.v1.logging.info('Beginning evaluation...')
    # Use the checkpointer class.
    self._checkpointer = checkpointer.Checkpointer(
        self._checkpoint_dir, self._checkpoint_file_prefix)
    checkpoint_version = -1
    # Check new checkpoints in a loop.
    while True:
      # Check if checkpoint exists.
      # Note that the existence of checkpoint 0 means that we have finished
      # iteration 0 (so we will start from iteration 1).
      latest_checkpoint_version = checkpointer.get_latest_checkpoint_number(
          self._checkpoint_dir)
      # checkpoints_iterator already makes sure a new checkpoint exists.
      if latest_checkpoint_version <= checkpoint_version:
        time.sleep(self._min_interval_secs)
        continue
      checkpoint_version = latest_checkpoint_version
      experiment_data = self._checkpointer.load_checkpoint(
          latest_checkpoint_version)
      assert self._agent.unbundle(self._checkpoint_dir,
                                  latest_checkpoint_version, experiment_data)

      self._run_eval_phase(experiment_data['total_steps'])
      if self._test_mode:
        break

  def _run_eval_phase(self, total_steps):
    """Runs evaluation phase given model has been trained for total_steps."""

    self._env.reset_sampler()
    self._initialize_metrics()

    num_episodes = 0
    episode_rewards = []

    while num_episodes < self._max_eval_episodes:
      _, episode_reward = self._run_one_episode()
      episode_rewards.append(episode_reward)
      num_episodes += 1

    self._write_metrics(total_steps, suffix='eval')

    output_file = os.path.join(self._output_dir, 'returns_%s' % total_steps)
    tf.compat.v1.logging.info('eval_file: %s', output_file)
    with tf.io.gfile.GFile(output_file, 'w+') as f:
      f.write(str(episode_rewards))



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class to represent the environment in the recommender system setting.

   Thus, it models things such as (1) the user's state, for example his/her
   interests and circumstances, (2) the documents available to suggest from and
   their properties, (3) simulates the selection of an item in the slate (or a
   no-op/quit), and (4) models the change in a user's state based on the slate
   presented and the document selected.

   The agent interacting with the environment is the recommender system.  The
   agent receives the state, which is an observation of the user's state and
   observations of the candidate documents. The agent then provides an action,
   which is a slate (an array of indices into the candidate set).

   The goal of the agent is to learn a recommendation policy: a policy that
   serves the user a slate (action) based on user and document features (state)
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
from recsim import document


class Environment(object):
  """Class to represent the environment in the recommender system simulator.

  Attributes:
    user_model: An instantiation of AbstractUserModel that represents a user.
    document_sampler: An instantiation of AbstractDocumentSampler.
    num_candidates: An integer representing the size of the candidate_set.
    slate_size: An integer representing the slate size.
    candidate_set: An instantiation of CandidateSet.
    num_clusters: An integer representing the number of document clusters.
  """

  def __init__(self,
               user_model,
               document_sampler,
               num_candidates,
               slate_size,
               resample_documents=True):
    """Initializes a new simulation environment.

    Args:
      user_model: An instantiation of AbstractUserModel
      document_sampler: An instantiation of AbstractDocumentSampler
      num_candidates: An integer representing the size of the candidate_set
      slate_size: An integer representing the slate size
      resample_documents: A boolean indicating whether to resample the candidate
        set every step
    """
    self._user_model = user_model
    self._document_sampler = document_sampler
    self._slate_size = slate_size
    self._num_candidates = num_candidates
    self._resample_documents = resample_documents

    # Create a candidate set.
    self._do_resample_documents()
    assert (slate_size <= num_candidates
           ), 'Slate size %d cannot be larger than number of candidates %d' % (
               slate_size, num_candidates)

  def _do_resample_documents(self):
    # TODO(sanmit): eventually model this creation with content creators.
    self._candidate_set = document.CandidateSet()
    for _ in range(self._num_candidates):
      self._candidate_set.add_document(self._document_sampler.sample_document())

  def reset(self):
    """Resets the environment and return the first observation.

    Returns:
      user_obs: An array of floats representing observations of the user's
        current state
      doc_obs: An OrderedDict of document observations keyed by document ids
    """
    self._user_model.reset()
    user_obs = self._user_model.create_observation()
    if self._resample_documents:
      self._do_resample_documents()
    self._current_documents = collections.OrderedDict(
        self._candidate_set.create_observation())
    return (user_obs, self._current_documents)

  def reset_sampler(self):
    self._document_sampler.reset_sampler()
    self._user_model.reset_sampler()

  @property
  def num_candidates(self):
    return self._num_candidates

  @property
  def slate_size(self):
    return self._slate_size

  @property
  def candidate_set(self):
    return self._candidate_set

  @property
  def user_model(self):
    return self._user_model

  def step(self, slate):
    """Executes the action, returns next state observation and reward.

    Args:
      slate: An integer array of size slate_size, where each element is an index
        into the set of current_documents presented

    Returns:
      user_obs: A gym observation representing the user's next state
      doc_obs: A list of observations of the documents
      responses: A list of AbstractResponse objects for each item in the slate
      done: A boolean indicating whether the episode has terminated
    """

    assert (len(slate) == self._slate_size
           ), 'Received unexpected slate size: expecting %s, got %s' % (
               self._slate_size, len(slate))

    # Get the documents associated with the slate
    doc_ids = list(self._current_documents)
    mapped_slate = [doc_ids[x] for x in slate]
    documents = self._candidate_set.get_documents(mapped_slate)
    # Simulate the user's response
    responses = self._user_model.simulate_response(documents)

    # Update the user's state.
    self._user_model.update_state(documents, responses)

    # Obtain next user state observation.
    user_obs = self._user_model.create_observation()

    # Check if reaches a terminal state and return.
    done = self._user_model.is_terminal()

    # Optionally, recreate the candidate set to simulate candidate
    # generators for the next query.
    if self._resample_documents:
      self._do_resample_documents()

    # Create observation of candidate set.
    self._current_documents = self._candidate_set.create_observation()

    return (user_obs, self._current_documents, responses, done)



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Agent that picks items with highest pCTR given the true user choice model."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from absl import logging

import numpy as np

from recsim import agent
from recsim import choice_model as cm


class GreedyPCTRAgent(agent.AbstractEpisodicRecommenderAgent):
  """An agent that recommends slates with the highest pCTR items.

  This agent assumes knowledge of the true underlying choice model. Note that
  this implicitly means it receives observations of the true user and document
  states. This agent myopically creates slates with items that have the highest
  probability of being clicked under the given choice model.
  """

  def __init__(self,
               action_space,
               belief_state,
               choice_model=cm.MultinomialLogitChoiceModel({'no_click_mass': 5
                                                           })):
    """Initializes a new greedy pCTR agent.

    Args:
      action_space: A gym.spaces object that specifies the format of actions
      belief_state: An instantiation of AbstractUserState assumed by the agent
      choice_model: An instantiation of AbstractChoiceModel assumed by the agent
        Default to a multinomial logit choice model with no_click_mass = 5.
    """

    super(GreedyPCTRAgent, self).__init__(action_space)
    self._choice_model = choice_model
    self._belief_state = belief_state

  def step(self, reward, observation):
    """Records the most recent transition and returns the agent's next action.

    We store the observation of the last time step since we want to store it
    with the reward.

    Args:
      reward: Unused.
      observation: A dictionary that includes the most recent observations and
        should have the following fields:
        - user: A list of floats representing the user's observed state
        - doc: A list of observations of document features

    Returns:
      slate: An integer array of size _slate_size, where each element is an
        index into the list of doc_obs
    """
    del reward  # Unused argument.
    doc_obs = observation['doc']

    # Score the documents without knowing the latent user state.
    self._choice_model.score_documents(self._belief_state, doc_obs.values())

    # Find the indices of the top scoring documents
    slate = self.findBestDocuments(self._choice_model.scores)

    # Return the slate of those documents
    logging.debug('Recommended slate: %s', slate)
    return slate

  def findBestDocuments(self, scores):
    """Returns the indices of the highest scores in sorted order.

    Args:
      scores: A list of floats representing unnormalized document scores

    Returns:
      sorted_indices: A list of integers indexing the highest scores, in sorted
      order
    """
    # Chose the k = slate_size best ones
    scores = np.array(scores)
    indices = np.argpartition(scores, -self._slate_size)[-self._slate_size:]

    # Sort them so the best appear first
    sorted_indices = indices[np.argsort(-scores[indices])]
    return sorted_indices



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper class to collect cluster click and impression counts."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from gym import spaces
import numpy as np

from recsim.agents.layers import sufficient_statistics


class ClusterClickStatsLayer(sufficient_statistics.SufficientStatisticsLayer):
  """Track impressions and clicks on a per-cluster basis and pass down to agent.

  This module assumes each document belongs to single cluster and we know the
  number of possible clusters. Every time we increase impression count for a
  cluster if the agent recommends a document from that cluster. We also increase
  click count for a cluster if user responds a click.
  """

  def __init__(self, base_agent_ctor, observation_space, action_space,
               **kwargs):
    """Initializes a ClusterClickStatsLayer object.

    Args:
      base_agent_ctor: a constructor for the base agent.
      observation_space: a gym.spaces object specifying the format of
        observations.
      action_space: A gym.spaces object that specifies the format of actions.
      **kwargs: arguments to pass to the downstream agent at construction time.
    """
    single_response_space = observation_space.spaces['response'].spaces[0]
    if 'cluster_id' not in single_response_space.spaces:
      raise ValueError('observation_space.spaces[\'response\']'
                       ' must contain \'cluster_id\' key.')
    cluster_id_space = single_response_space.spaces['cluster_id']
    if isinstance(cluster_id_space, spaces.Box):
      if len(cluster_id_space.high) > 1:
        raise ValueError('cluster_id response field must be 0 dimensional.')
      num_clusters = cluster_id_space.high
    elif isinstance(cluster_id_space, spaces.Discrete):
      num_clusters = cluster_id_space.n
    else:
      raise ValueError('cluster_id response field must be either gym.spaces.Box'
                       ' or gym spaces.Discrete')
    self._num_clusters = num_clusters
    if 'click' not in single_response_space.spaces:
      raise ValueError(
          'observation_space.spaces[\'response\'] must contain \'click\' key.')
    suf_stat_space = spaces.Dict({
        'impression_count':
            spaces.Box(
                shape=(num_clusters,), dtype=np.float32, low=0.0, high=np.inf),
        'click_count':
            spaces.Box(
                shape=(num_clusters,), dtype=np.float32, low=0.0, high=np.inf)
    })
    super(ClusterClickStatsLayer,
          self).__init__(base_agent_ctor, observation_space, action_space,
                         suf_stat_space, **kwargs)

  def _create_observation(self):
    return {
        'impression_count':
            np.array(self._sufficient_statistics['impression_count']),
        'click_count':
            np.array(self._sufficient_statistics['click_count']),
    }

  def _update(self, observation):
    """Updates user impression/click count given user response on each item."""
    if self._sufficient_statistics is None:
      self._sufficient_statistics = {
          'impression_count': [
              0,
          ] * self._num_clusters,
          'click_count': [
              0,
          ] * self._num_clusters
      }
    if observation['response'] is not None:
      for response in observation['response']:
        cluster_id = int(response['cluster_id'])
        self._sufficient_statistics['impression_count'][cluster_id] += 1
        if response['click']:
          self._sufficient_statistics['click_count'][cluster_id] += 1



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""An example of main function for training in RecSim.

Use the interest evolution environment and a slateQ agent for illustration.

To run locally:

python main.py --base_dir=/tmp/interest_evolution \
  --gin_bindings=simulator.runner_lib.Runner.max_steps_per_episode=50

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import app
from absl import flags
import numpy as np
from recsim.agents import full_slate_q_agent
from recsim.environments import interest_evolution
from recsim.simulator import runner_lib


FLAGS = flags.FLAGS


def create_agent(sess, environment, eval_mode, summary_writer=None):
  """Creates an instance of FullSlateQAgent.

  Args:
    sess: A `tf.Session` object for running associated ops.
    environment: A recsim Gym environment.
    eval_mode: A bool for whether the agent is in training or evaluation mode.
    summary_writer: A Tensorflow summary writer to pass to the agent for
      in-agent training statistics in Tensorboard.

  Returns:
    An instance of FullSlateQAgent.
  """
  kwargs = {
      'observation_space': environment.observation_space,
      'action_space': environment.action_space,
      'summary_writer': summary_writer,
      'eval_mode': eval_mode,
  }
  return full_slate_q_agent.FullSlateQAgent(sess, **kwargs)


def main(argv):
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')

  runner_lib.load_gin_configs(FLAGS.gin_files, FLAGS.gin_bindings)
  seed = 0
  slate_size = 2
  np.random.seed(seed)
  env_config = {
      'num_candidates': 5,
      'slate_size': slate_size,
      'resample_documents': True,
      'seed': seed,
  }

  runner = runner_lib.TrainRunner(
      base_dir=FLAGS.base_dir,
      create_agent_fn=create_agent,
      env=interest_evolution.create_environment(env_config),
      episode_log_file=FLAGS.episode_log_file,
      max_training_steps=50,
      num_iterations=10)
  runner.run_experiment()

  runner = runner_lib.EvalRunner(
      base_dir=FLAGS.base_dir,
      create_agent_fn=create_agent,
      env=interest_evolution.create_environment(env_config),
      max_eval_episodes=5,
      test_mode=True)
  runner.run_experiment()


if __name__ == '__main__':
  flags.mark_flag_as_required('base_dir')
  app.run(main)



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A wrapper for using Gym environment."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import gym
from gym import spaces
import numpy as np


def _dummy_metrics_aggregator(responses, metrics, info):
  del responses  # Unused.
  del metrics  # Unused.
  del info  # Unused.
  return


def _dummy_metrics_writer(metrics, add_summary_fn):
  del metrics  # Unused.
  del add_summary_fn  # Unused.
  return


class RecSimGymEnv(gym.Env):
  """Class to wrap recommender system environment to gym.Env.

  Attributes:
    game_over: A boolean indicating whether the current game has finished
    action_space: A gym.spaces object that specifies the space for possible
      actions.
    observation_space: A gym.spaces object that specifies the space for possible
      observations.
  """

  def __init__(self,
               raw_environment,
               reward_aggregator,
               metrics_aggregator=_dummy_metrics_aggregator,
               metrics_writer=_dummy_metrics_writer):
    """Initializes a RecSim environment conforming to gym.Env.

    Args:
      raw_environment: A recsim recommender system environment.
      reward_aggregator: A function mapping a list of responses to a number.
      metrics_aggregator: A function aggregating metrics over all steps given
        responses and response_names.
      metrics_writer:  A function writing final metrics to TensorBoard.
    """
    self._environment = raw_environment
    self._reward_aggregator = reward_aggregator
    self._metrics_aggregator = metrics_aggregator
    self._metrics_writer = metrics_writer
    self.reset_metrics()

  @property
  def environment(self):
    """Returns the recsim recommender system environment."""
    return self._environment

  @property
  def game_over(self):
    return False

  @property
  def action_space(self):
    """Returns the action space of the environment.

    Each action is a vector that specified document slate. Each element in the
    vector corresponds to the index of the document in the candidate set.
    """
    return spaces.MultiDiscrete(self._environment.num_candidates * np.ones(
        (self._environment.slate_size,)))

  @property
  def observation_space(self):
    """Returns the observation space of the environment.

    Each observation is a dictionary with three keys `user`, `doc` and
    `response` that includes observation about user state, document and user
    response, respectively.
    """
    return spaces.Dict({
        'user': self._environment.user_model.observation_space(),
        'doc': self._environment.candidate_set.observation_space(),
        'response': self._environment.user_model.response_space(),
    })

  def step(self, action):
    """Runs one timestep of the environment's dynamics.

    When end of episode is reached, you are responsible for calling `reset()`
    to reset this environment's state. Accepts an action and returns a tuple
    (observation, reward, done, info).

    Args:
      action (object): An action provided by the environment

    Returns:
      A four-tuple of (observation, reward, done, info) where:
        observation (object): agent's observation that include
          1. User's state features
          2. Document's observation
          3. Observation about user's slate responses.
        reward (float) : The amount of reward returned after previous action
        done (boolean): Whether the episode has ended, in which case further
          step() calls will return undefined results
        info (dict): Contains responses for the full slate for
          debugging/learning.
    """
    user_obs, doc_obs, responses, done = self._environment.step(action)
    obs = dict(
        user=user_obs,
        doc=doc_obs,
        response=tuple(
            response.create_observation() for response in responses))

    # extract rewards from responses
    reward = self._reward_aggregator(responses)
    info = self.extract_env_info()
    return obs, reward, done, info

  def reset(self):
    user_obs, doc_obs = self._environment.reset()
    return dict(user=user_obs, doc=doc_obs, response=None)

  def reset_sampler(self):
    self._environment.reset_sampler()

  def render(self, mode='human'):
    raise NotImplementedError

  def close(self):
    raise NotImplementedError

  def seed(self, seed=None):
    np.random.seed(seed=seed)

  def extract_env_info(self):
    info = {'env': self._environment}
    return info

  def reset_metrics(self):
    """Resets every metric to zero.

    We reset metrics for every iteration but not every episode. On the other
    hand, reset() gets called for every episode.
    """
    self._metrics = collections.defaultdict(float)

  def update_metrics(self, responses, info=None):
    """Updates metrics with one step responses."""
    self._metrics = self._metrics_aggregator(
        responses, self._metrics, info)

  def write_metrics(self, add_summary_fn):
    """Writes metrics to TensorBoard by calling add_summary_fn."""
    self._metrics_writer(self._metrics, add_summary_fn)



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for recsim.agents.greedy_pctr_agent."""

from gym import spaces
import numpy as np

from recsim.agents import greedy_pctr_agent
from recsim.environments import interest_exploration as ie
from recsim.simulator import environment
import tensorflow as tf


class GreedyPCTRAgentTest(tf.test.TestCase):

  def test_find_best_documents(self):
    action_space = spaces.MultiDiscrete(4 * np.ones((4,)))
    agent = greedy_pctr_agent.GreedyPCTRAgent(action_space, None)
    scores = [-1, -2, 4.32, 0, 15, -6, 4.32]
    indices = agent.findBestDocuments(scores)
    self.assertAllEqual(indices, [4, 2, 6, 3])

  def test_step(self):
    # Create a simple user
    slate_size = 2
    num_candidates = 5
    action_space = spaces.MultiDiscrete(num_candidates * np.ones((slate_size,)))
    user_model = ie.IEUserModel(
        slate_size,
        user_state_ctor=ie.IEUserState,
        response_model_ctor=ie.IEResponse)

    # Create a set of documents
    document_sampler = ie.IETopicDocumentSampler(seed=1)
    ieenv = environment.Environment(
        user_model,
        document_sampler,
        num_candidates,
        slate_size,
        resample_documents=True)

    # Create agent
    agent = greedy_pctr_agent.GreedyPCTRAgent(action_space,
                                              user_model.avg_user_state)

    # This agent doesn't use the previous user response
    observation, documents = ieenv.reset()
    slate = agent.step(1, dict(user=observation, doc=documents))
    scores = [
        user_model.avg_user_state.score_document(doc_obs)
        for doc_obs in list(documents.values())
    ]
    expected_slate = sorted(np.argsort(scores)[-2:])
    self.assertAllEqual(sorted(slate), expected_slate)

  def test_bundle_and_unbundle_trivial(self):
    action_space = spaces.MultiDiscrete(np.ones((1,)))
    agent = greedy_pctr_agent.GreedyPCTRAgent(action_space, None)
    self.assertFalse(agent.unbundle('', 0, {}))
    self.assertEqual({
        'episode_num': 0
    }, agent.bundle_and_checkpoint('', 0))

  def test_bundle_and_unbundle(self):
    # Initialize agent
    slate_size = 1
    num_candidates = 3
    action_space = spaces.MultiDiscrete(num_candidates * np.ones((slate_size,)))

    user_model = ie.IEUserModel(
        slate_size,
        user_state_ctor=ie.IEUserState,
        response_model_ctor=ie.IEResponse)
    agent = greedy_pctr_agent.GreedyPCTRAgent(action_space,
                                              user_model.avg_user_state)

    # Create a set of documents
    document_sampler = ie.IETopicDocumentSampler()
    documents = {}
    for i in range(num_candidates):
      video = document_sampler.sample_document()
      documents[i] = video.create_observation()

    # Test that slate indices in correct range and length is correct
    observation = dict(user=user_model.create_observation(), doc=documents)
    agent.step(1, observation)

    bundle_dict = agent.bundle_and_checkpoint('', 0)
    self.assertTrue(agent.unbundle('', 0, bundle_dict))
    self.assertEqual(bundle_dict, agent.bundle_and_checkpoint('', 0))


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for recsim.agents.ucb1_agent."""

import numpy as np
from recsim.agents import cluster_bandit_agent
from recsim.agents.layers import cluster_click_statistics
from recsim.environments import interest_exploration as ie
import tensorflow as tf


class UCB1AgentTest(tf.test.TestCase):

  def test_step(self):
    # Initialize agent.
    env_config = {
        'slate_size': 1,
        'num_candidates': 5,
        'resample_documents': True,
        'seed': 1,
    }
    env = ie.create_environment(env_config)
    kwargs = {
        'observation_space': env.observation_space,
        'action_space': env.action_space,
    }
    agent = cluster_click_statistics.ClusterClickStatsLayer(
        cluster_bandit_agent.ClusterBanditAgent, **kwargs)

    observation1, documents1 = env.environment.reset()
    slate1 = agent.step(0,
                        dict(user=observation1, doc=documents1, response=None))
    # Pick the document with the best quality in Topic 0.
    scores_c0 = [(features['quality'] if features['cluster_id'] == 0 else 0)
                 for _, features in documents1.items()]
    scores_c1 = [(features['quality'] if features['cluster_id'] == 1 else 0)
                 for _, features in documents1.items()]
    self.assertIn(slate1[0], [np.argmax(scores_c0), np.argmax(scores_c1)])
    picked_cluster = list(documents1.values())[slate1[0]]['cluster_id']

    observation2, documents, response1, _ = env.environment.step(slate1)
    response1_obs = [response.create_observation() for response in response1]
    response1_obs[0]['cluster_id'] = picked_cluster
    slate2 = agent.step(
        ie.total_clicks_reward(response1),
        dict(user=observation2, doc=documents, response=response1_obs))
    # Pick Topic 1 because we have no observation about it.
    # Pick the document with the best quality there.
    doc_qualities = [
        (features['quality'] if features['cluster_id'] != picked_cluster else 0)
        for _, features in documents.items()
    ]
    self.assertAllEqual(slate2, [
        np.argmax(doc_qualities),
    ])

    self.assertNotEqual(
        list(documents.values())[slate2[0]]['cluster_id'], picked_cluster)

    observation3, documents, response2, _ = env.environment.step(slate2)
    response2_obs = [response.create_observation() for response in response2]
    # Make a clicked response.
    response2_obs[0]['click'] = 1
    response2_obs[0]['cluster_id'] = 1 - picked_cluster
    slate3 = agent.step(
        ie.total_clicks_reward(response2),
        dict(user=observation3, doc=documents, response=response2_obs))
    # Pick the first topic which has the best UCB and then pick the document
    # with the best quality in it.
    pulls = np.array([1, 1], dtype=np.float)
    rewards = np.array([0, 0], dtype=np.float)
    rewards[1 - picked_cluster] = 1
    ct = np.sqrt(2.0 * np.log(2.0))
    topic_index = rewards / pulls + ct * np.sqrt(1.0 / pulls)
    doc_qualities = [(features['quality'] if
                      features['cluster_id'] == np.argmax(topic_index) else 0)
                     for _, features in documents.items()]
    self.assertAllEqual(slate3, [np.argmax(doc_qualities)])

    agent.end_episode(
        ie.total_clicks_reward(response2),
        dict(user=observation3, doc=documents, response=response2_obs))
    slate4 = agent.step(0,
                        dict(user=observation1, doc=documents1, response=None))
    self.assertAllEqual(slate4, slate1)


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for recsim.agents.tabular_q_agent."""

from gym import spaces
import numpy as np
from recsim.agents import tabular_q_agent
from recsim.testing import test_environment as te
import tensorflow as tf


class TabularQAgentTest(tf.test.TestCase):

  def init_agent_and_env(self,
                         slate_size=1,
                         num_candidates=10,
                         learning_rate=0.8,
                         gamma=0.0,
                         policy='epsilon_greedy',
                         starting_probs=(1.0, 0.0, 0.0, 0.0, 0.0, 0.0)):
    env_config = {
        'num_candidates': num_candidates,
        'slate_size': slate_size,
        'resample_documents': False,
        'seed': 42,
        'starting_probs': starting_probs
    }
    te_sim = te.create_environment(env_config)
    agent = tabular_q_agent.TabularQAgent(
        te_sim.observation_space,
        te_sim.action_space,
        gamma=gamma,
        exploration_policy=policy,
        learning_rate=learning_rate)
    return te_sim, agent

  def test_step(self):
    te_sim, agent = self.init_agent_and_env()
    observation0 = te_sim.reset()
    slate1 = agent.step(0, observation0)
    selected_doc0 = list(observation0['doc'].values())[slate1[0]]
    # Environment always starts at state 0.
    self.assertEqual(agent._previous_state_action_index, (selected_doc0, 0))
    observation1, reward1, _, _ = te_sim.step(slate1)
    slate2 = agent.step(reward1, observation1)
    selected_doc1 = list(observation1['doc'].values())[slate2[0]]
    observed_state = observation1['user']
    self.assertEqual(agent._previous_state_action_index,
                     (selected_doc1, observed_state))
    self.assertEqual(agent._q_value_table,
                     {(selected_doc0, 0): agent._learning_rate * -10.0})
    self.assertEqual(agent._state_action_counts, {(selected_doc0, 0): 1})

  def test_myopic_value_estimation(self):
    te_sim, agent = self.init_agent_and_env()
    observation0 = te_sim.reset()
    slate = agent.step(0, observation0)
    for _ in range(5000):
      observation, reward, _, _ = te_sim.step(slate)
      slate = agent.step(reward, observation)
    for state in range(6):
      for action in range(4):
        self.assertAlmostEqual(agent._q_value_table[(action, state)],
                               te.QVALUES0[state][action])

  def test_gamma05_value_estimation(self):
    # TODO(mmladenov): tune further to improve speed of this test
    te_sim, agent = self.init_agent_and_env(gamma=0.5)
    observation = te_sim.reset()
    reward = 0
    for i in range(100, 100100):
      slate = agent.step(reward, observation)
      observation, reward, _, _ = te_sim.step(slate)
      agent._learning_rate = 100.0 / float(i)
    for state in range(6):
      for action in range(4):
        self.assertAlmostEqual(
            agent._q_value_table[(action, state)],
            te.QVALUES05[state][action],
            delta=0.2)

  def test_dicretize_gym_leaf(self):
    _, agent = self.init_agent_and_env()
    self.assertEqual(
        agent._discretize_gym_leaf(spaces.Discrete(5), [
            4,
        ]), [
            4,
        ])
    box = spaces.Box(
        low=agent._discretization_bins[0],
        high=agent._discretization_bins[-1],
        shape=(1, 1),
        dtype=np.float32)
    # Some corner cases in 1d and 2x2d.
    self.assertEqual(
        agent._discretize_gym_leaf(box, [
            agent._discretization_bins[0] - 10E-5,
        ]), [
            0,
        ])
    self.assertEqual(
        agent._discretize_gym_leaf(box, [
            agent._discretization_bins[0],
        ]), [
            1,
        ])
    self.assertEqual(
        agent._discretize_gym_leaf(box, [
            agent._discretization_bins[-1] - 10E-5,
        ]), [
            len(agent._discretization_bins) - 1,
        ])
    self.assertEqual(
        agent._discretize_gym_leaf(box, [
            agent._discretization_bins[-1] + 1.0,
        ]), [
            len(agent._discretization_bins),
        ])
    box2x2 = spaces.Box(
        low=agent._discretization_bins[0],
        high=agent._discretization_bins[-1],
        shape=(2, 2),
        dtype=np.float32)
    self.assertEqual(
        agent._discretize_gym_leaf(box2x2, [
            np.array([[
                agent._discretization_bins[0] - 10E-6,
                agent._discretization_bins[-1] - 10E-6
            ],
                      [
                          agent._discretization_bins[-1] + 1.0,
                          agent._discretization_bins[0]
                      ]]),
        ]), [0, 99, 100, 1])

  def test_slate_enumeration(self):
    te_sim, agent = self.init_agent_and_env(slate_size=2, num_candidates=4)
    observation0 = te_sim.reset()
    slates = [(0, 1), (0, 2), (0, 3), (1, 2), (1, 3), (2, 3)]
    enumerated_slates = [
        slate for slate, _ in agent._enumerate_slates(observation0['doc'])
    ]
    self.assertCountEqual(slates, enumerated_slates)

  def test_bundle_and_unbundle(self):
    te_sim, agent = self.init_agent_and_env(
        slate_size=1, num_candidates=4, policy='min_count')
    # Make a few steps to populate counts and Q-table.
    observation0 = te_sim.reset()
    slate1 = agent.step(0, observation0)
    observation1, reward1, _, _ = te_sim.step(slate1)
    agent.step(reward1, observation1)
    bundle_dict = {
        'q_value_table': agent._q_value_table,
        'sa_count': agent._state_action_counts
    }
    self.assertEqual(bundle_dict, agent.bundle_and_checkpoint('', 0))
    _, new_agent = self.init_agent_and_env(slate_size=1, num_candidates=4)
    self.assertTrue(new_agent.unbundle('', 0, bundle_dict))
    self.assertEqual(bundle_dict['q_value_table'], new_agent._q_value_table)
    self.assertEqual(bundle_dict['sa_count'], new_agent._state_action_counts)


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Agent that implements the Slate-Q algorithms."""

import itertools
import gin.tf
from gym import spaces
from recsim import agent as abstract_agent
from recsim.agents.dopamine import dqn_agent
import tensorflow as tf


@gin.configurable
class FullSlateQAgent(dqn_agent.DQNAgentRecSim,
                      abstract_agent.AbstractEpisodicRecommenderAgent):
  """A recommender agent implements full slate Q-learning based on DQN agent.

  This is a standard, nondecomposed Q-learning method that treats each slate
  atomically (i.e., holistically) as a single action.
  """

  def __init__(self,
               sess,
               observation_space,
               action_space,
               optimizer_name='',
               eval_mode=False,
               **kwargs):
    """Initializes a FullSlateQAgent.

    Args:
      sess: a Tensorflow session.
      observation_space: A gym.spaces object that specifies the format of
        observations.
      action_space: A gym.spaces object that specifies the format of actions.
      optimizer_name: The name of the optimizer.
      eval_mode: A bool for whether the agent is in training or evaluation mode.
      **kwargs: Keyword arguments to the DQNAgent.
    """
    self._num_candidates = int(action_space.nvec[0])
    abstract_agent.AbstractEpisodicRecommenderAgent.__init__(self, action_space)
    # Each slate is a single action. Assume ordering of items matters.
    self._all_possible_slates = [
        x for x in itertools.permutations(
            range(self._num_candidates), action_space.nvec.shape[0])
    ]
    num_actions = len(self._all_possible_slates)
    self._env_action_space = spaces.Discrete(num_actions)

    dqn_agent.DQNAgentRecSim.__init__(
        self,
        sess,
        observation_space,
        num_actions=num_actions,
        stack_size=1,
        optimizer_name='',
        eval_mode=eval_mode,
        **kwargs)

  # Builds a tower to compute Q-value for each possible slate.
  def _network_adapter(self, states, scope):
    self._validate_states(states)

    with tf.compat.v1.name_scope('network'):
      q_value_list = []
      for slate in self._all_possible_slates:
        user = tf.squeeze(states[:, 0, :, :], axis=2)
        docs = []
        for i in slate:
          docs.append(tf.squeeze(states[:, i + 1, :, :], axis=2))
        q_value_list.append(self.network(user, tf.concat(docs, axis=1), scope))
      q_values = tf.concat(q_value_list, axis=1)

    return dqn_agent.DQNNetworkType(q_values)

  def _build_networks(self):
    with tf.compat.v1.name_scope('networks'):
      self._replay_net_outputs = self._network_adapter(self._replay.states,
                                                       'Online')
      self._replay_next_target_net_outputs = self._network_adapter(
          self._replay.states, 'Target')
      self._net_outputs = self._network_adapter(self.state_ph, 'Online')
      self._q_argmax = tf.argmax(input=self._net_outputs.q_values, axis=1)[0]

  def step(self, reward, observation):
    """Receives observations of environment and returns a slate.

    Args:
      reward: A double representing the overall reward to the recommended slate.
      observation: A dictionary that stores all the observations including:
        - user: A list of floats representing the user's observed state
        - doc: A list of observations of document features
        - response: A vector valued response signal that represent user's
          response to each document

    Returns:
      slate: An integer array of size _slate_size, where each element is an
        index in the list of document observvations.
    """
    return self._all_possible_slates[super(FullSlateQAgent, self).step(
        reward, self._obs_adapter.encode(observation))]

  def _build_replay_buffer(self, use_staging):
    """Creates the replay buffer used by the agent.

    Args:
      use_staging: bool, if True, uses a staging area to prefetch data for
        faster training.

    Returns:
      A WrapperReplayBuffer object.
    """
    return dqn_agent.wrapped_replay_buffer(
        observation_shape=self.observation_shape,
        stack_size=self.stack_size,
        use_staging=use_staging,
        update_horizon=self.update_horizon,
        gamma=self.gamma,
        observation_dtype=self.observation_dtype)

  def begin_episode(self, observation):
    """Returns the agent's first action for this episode.

    Args:
      observation: numpy array, the environment's initial observation.

    Returns:
      An integer array of size _slate_size, the selected slated, each
      element of which is an index in the list of doc_obs.
    """
    return self._all_possible_slates[super(FullSlateQAgent, self).begin_episode(
        self._obs_adapter.encode(observation))]

  def end_episode(self, reward, observation):
    """Signals the end of the episode to the agent.

    We store the observation of the current time step, which is the last
    observation of the episode.

    Args:
      reward: float, the last reward from the environment.
      observation: numpy array, the environment's initial observation.
    """
    super(FullSlateQAgent, self).end_episode(reward)



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Abstract classes that encode a user's state and dynamics."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
from gym import spaces
import numpy as np


class AbstractResponse(object):  # pytype: disable=ignored-metaclass
  """Abstract class to model a user response."""

  __metaclass__ = abc.ABCMeta

  @staticmethod
  @abc.abstractmethod
  def response_space():
    """ArraySpec that defines how a single response is represented."""

  @abc.abstractmethod
  def create_observation(self):
    """Creates a tensor observation of this response."""


class AbstractUserState(object):  # pytype: disable=ignored-metaclass
  """Abstract class to represent a user's state."""
  __metaclass__ = abc.ABCMeta

  # Number of features to represent the user's interests.
  NUM_FEATURES = None

  @abc.abstractmethod
  def create_observation(self):
    """Generates obs of underlying state to simulate partial observability.

    Returns:
      obs: A float array of the observed user features.
    """

  @staticmethod
  @abc.abstractmethod
  def observation_space():
    """Gym.spaces object that defines how user states are represented."""


class AbstractUserSampler(object):  # pytype: disable=ignored-metaclass
  """Abstract class to sample users."""

  __metaclass__ = abc.ABCMeta

  def __init__(self, user_ctor, seed=0):
    """Creates a new user state sampler.

    User states of the type user_ctor are sampled.

    Args:
      user_ctor: A class/constructor for the type of user states that will be
        sampled.
      seed: An integer for a random seed.
    """
    self._user_ctor = user_ctor
    self._seed = seed
    self.reset_sampler()

  def reset_sampler(self):
    self._rng = np.random.RandomState(self._seed)

  @abc.abstractmethod
  def sample_user(self):
    """Creates a new instantiation of this user's hidden state parameters."""

  def get_user_ctor(self):
    """Returns the constructor/class of the user states that will be sampled."""
    return self._user_ctor


class AbstractUserModel(object):  # pytype: disable=ignored-metaclass
  """Abstract class to represent an encoding of a user's dynamics."""

  __metaclass__ = abc.ABCMeta

  def __init__(self, response_model_ctor, user_sampler, slate_size):
    """Initializes a new user model.

    Args:
      response_model_ctor: A class/constructor representing the type of
        responses this model will generate.
      user_sampler: An instance of AbstractUserSampler that can generate
        initial user states from an inital state distribution.
      slate_size: integer number of documents that can be served to the user at
        any interaction.
    """
    if not response_model_ctor:
      raise TypeError('response_model_ctor is a required callable')

    self._user_sampler = user_sampler
    self._user_state = self._user_sampler.sample_user()
    self._response_model_ctor = response_model_ctor
    self._slate_size = slate_size

  ## Transition model
  @abc.abstractmethod
  def update_state(self, slate_documents, responses):
    """Updates the user's state based on the slate and document selected.

    Args:
      slate_documents: A list of AbstractDocuments for items in the slate.
      responses: A list of AbstractResponses for each item in the slate.
    Updates: The user's hidden state.
    """

  def reset(self):
    """Resets the user."""
    self._user_state = self._user_sampler.sample_user()

  def reset_sampler(self):
    """Resets the sampler."""
    self._user_sampler.reset_sampler()

  @abc.abstractmethod
  def is_terminal(self):
    """Returns a boolean indicating whether this session is over."""

  ## Choice model
  @abc.abstractmethod
  def simulate_response(self, documents):
    """Simulates the user's response to a slate of documents.

    This could involve simulating models of attention, as well as random
    sampling for selection from scored documents.

    Args:
      documents: a list of AbstractDocuments

    Returns:
      (response) a list of AbstractResponse objects for each slate item
    """

  def response_space(self):
    res_space = self._response_model_ctor.response_space()
    return spaces.Tuple(tuple([
        res_space,
    ] * self._slate_size))

  def get_response_model_ctor(self):
    """Returns a constructor for the type of response this model will create."""
    return self._response_model_ctor

  def observation_space(self):
    """A Gym.spaces object that describes possible user observations."""
    return self._user_state.observation_space()

  def create_observation(self):
    """Emits obesrvation about user's state."""
    return self._user_state.create_observation()



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper classe to record fixed length history of observations."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from gym import spaces

from recsim.agents.layers import sufficient_statistics


class FixedLengthHistoryLayer(sufficient_statistics.SufficientStatisticsLayer):
  r"""Creates a buffer of the last k rewards and observations.

  This module introduces sufficient statistics in the form of a buffer holding
  the last k (specified by history_length) observations. This buffer is injected
  into observation_space[\'user\'][\'sufficient_statistics\'] in the form of a
  gym.spaces.Tuple of length up to k . In the first k-1 steps of the episode
  there are not enough observations to fill the buffer, so they will be filled
  with None. Each non-vacuous element of the tuple is an instance of
  (a subset of) observation_space.
  """

  def __init__(self,
               base_agent_ctor,
               observation_space,
               action_space,
               history_length,
               remember_user=True,
               remember_response=True,
               remember_doc=False,
               **kwargs):
    r"""Initializes a FixedLengthHistoryLayer object.

    Args:
      base_agent_ctor: a constructor for the base agent.
      observation_space: a gym.spaces object specifying the format of
        observations.
      action_space: A gym.spaces object that specifies the format of actions.
      history_length: positive integer number of observations to remember.
      remember_user: boolean, indicates whether to track
        observation_space[\'user\'].
      remember_response: boolean, indicates whether to track
        observation_space[\'response\'].
      remember_doc: boolean, indicates whether to track
        observation_space[\'doc\'].
      **kwargs: arguments to pass to the downstream agent at construction time.
    """

    self._history_length = history_length
    self._features = []
    if remember_user:
      self._features.append('user')
    if remember_response:
      self._features.append('response')
    if remember_doc:
      self._features.append('doc')
    observation_space_to_remember = spaces.Dict({
        feature: observation_space[feature] for feature in self._features
    })
    suf_stat_space = spaces.Tuple([
        observation_space_to_remember,
    ] * history_length)
    super(FixedLengthHistoryLayer,
          self).__init__(base_agent_ctor, observation_space, action_space,
                         suf_stat_space, **kwargs)

  def _create_observation(self):
    return tuple(self._sufficient_statistics)

  def _update(self, observation):
    """Updates user impression/click count given user response on each item."""
    if self._sufficient_statistics is None:
      self._sufficient_statistics = self._history_length * [
          None,
      ]

    observation_to_remember = {
        feature: observation[feature] for feature in self._features
    }
    self._sufficient_statistics = [
        observation_to_remember,
    ] + self._sufficient_statistics[:-1]



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for recsim.utils."""

import collections
import numpy as np
from recsim import utils
import tensorflow as tf


class UtilsTest(tf.test.TestCase):

  def test_aggregate_video_cluster_metrics(self):
    metrics = collections.defaultdict(float)
    metrics['impression'] = 10
    metrics['cluster_watch_count_cluster_0'] = 1
    metrics['cluster_watch_count_no_click'] = 9
    metrics['quality'] = 0.7
    metrics['click'] = 1
    responses = ({
        'click': 1,
        'quality': np.array(0.5),
        'cluster_id': np.array(1)
    }, {
        'click': 0,
        'quality': np.array(0.8),
        'cluster_id': np.array(2)
    })
    metrics = utils.aggregate_video_cluster_metrics(responses, metrics)
    self.assertEqual(
        metrics, {
            'impression': 11.0,
            'cluster_watch_count_cluster_0': 1.0,
            'cluster_watch_count_cluster_1': 1.0,
            'cluster_watch_count_no_click': 9.0,
            'quality': 1.2,
            'click': 2.0
        })


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A Tabular Q-learning implementation."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import itertools

from absl import logging
from gym import spaces
import numpy as np

from recsim import agent
from recsim.agents import agent_utils


class TabularQAgent(agent.AbstractEpisodicRecommenderAgent):
  """Tabular Q-learning agent with universal function approximation.

  This agent provides a tabular implementation of the Q-learning algorithm.
  To construct a tabular representation of the state-action space, the agent
  does the following:
    1. the action space consists of all ordered k-tuples of document features
    available in observation['doc'];
    2. the state space consists of observation['user']
    and observation['response'];
    3. the observation and action space are joined and flattened;
    4. all continuoius values in the flattened state-action vector are
    discretized into a predefined number of bins.
  In the tabularized state-action space, the agent applies the standard
  Q-learning update of
    Q_{t+1}(s,a) = (1-a) * Q_t(s,a) + a * (R(s,a) + g * max_a(Q_t(s', a))).
  Assuming that the discretization of countinous attributes is fine enough, and
  the problem itself is Markovian given the observations, the output of this
  agent can be assumed to converge to a close approximation of the ground truth
  Q-function. Producing ground truth Q-functions is the main intended use of
  this agent, since discretization is prohibitively expensive in
  high-dimensional environments.
  """

  def __init__(self,
               observation_space,
               action_space,
               eval_mode=False,
               ignore_response=True,
               discretization_bounds=(0.0, 10.0),
               number_bins=100,
               exploration_policy='epsilon_greedy',
               exploration_temperature=0.99,
               learning_rate=0.1,
               gamma=0.99,
               **kwargs):
    """TabularQAgent init.

    Args:
      observation_space: a gym.spaces object specifying the format of
        observations.
      action_space: a gym.spaces object that specifies the format of actions.
      eval_mode: Boolean indicating whether the agent is in training or eval
        mode.
      ignore_response: Boolean indicating whether the agent should ignore the
        response part of the observation.
      discretization_bounds: pair of real numbers indicating the min and max
        value for continuous attributes discretization. Values below the min
        will all be grouped in the first bin, while values above the max will
        all be grouped in the last bin. See the documentation of numpy.digitize
        for further details.
      number_bins: positive integer number of bins used to discretize continuous
        attributes.
      exploration_policy: either one of ['epsilon_greedy', 'min_count'] or a
        custom function. TODO(mmladenov): formalize requirements of this
          function.
      exploration_temperature: a real number passed as parameter to the
        exploration policy.
      learning_rate: a real number between 0 and 1 indicating how much to update
        Q-values, i.e. Q_t+1(s,a) = (1 - learning_rate) * Q_t(s, a)
                                     + learning_rate * (R(s,a) + ...).
      gamma: real value between 0 and 1 indicating the discount factor of the
        MDP.
      **kwargs: additional arguments like eval_mode.
    """
    self._kwargs = kwargs
    super(TabularQAgent, self).__init__(action_space)
    # hard params
    self._gamma = gamma
    self._eval_mode = eval_mode
    self._previous_slate = None
    self._learning_rate = learning_rate
    # storage
    self._q_value_table = {}
    self._state_action_counts = {}
    self._previous_state_action_index = None
    # discretization and spaces
    self._discretization_bins = np.linspace(
        discretization_bounds[0], discretization_bounds[1], num=number_bins)
    single_doc_space = list(observation_space.spaces['doc'].spaces.values())[0]
    slate_tuple = tuple([single_doc_space] * self._slate_size)
    action_space = spaces.Tuple(slate_tuple)
    self._ignore_response = ignore_response
    state_action_space = {
        'user': observation_space.spaces['user'],
        'action': action_space
    }
    if not self._ignore_response:
      state_action_space['response'] = observation_space.spaces['response']
    self._state_action_space = spaces.Dict(state_action_space)
    self._observation_featurizer = agent_utils.GymSpaceWalker(
        self._state_action_space, self._discretize_gym_leaf)
    # exploration
    self._exploration_policy = exploration_policy
    self._exploration_temperature = exploration_temperature
    self._base_exploration_temperature = self._exploration_temperature
    self._exploration_functions = {
        'epsilon_greedy':
            lambda observation: agent_utils.epsilon_greedy_exploration(  # pylint: disable=g-long-lambda
                self._enumerate_state_action_indices(observation), self.
                _q_value_table, self._exploration_temperature),
        'min_count':
            lambda observation: agent_utils.min_count_exploration(  # pylint: disable=g-long-lambda
                self._enumerate_state_action_indices(observation),
                self._state_action_counts)
    }

  def _discretize_gym_leaf(self, gym_space, gym_observations):

    index = []
    for gym_observation in gym_observations:
      gym_observation = gym_observations[0]
      if isinstance(gym_space, spaces.box.Box):
        gym_observation = np.array(gym_observation)
        dis_obs = np.digitize(gym_observation.flatten(),
                              self._discretization_bins)
        index += list(dis_obs)
      elif isinstance(gym_space, spaces.discrete.Discrete):
        index.append(gym_observation)
      else:
        raise NotImplementedError('Gym space type ' + str(type(gym_space)) +
                                  ' not implemented yet.')
    return index

  def _enumerate_slates(self, doc_dict):
    documents = list(doc_dict.values())
    num_documents = len(documents)
    for slate in itertools.combinations(range(num_documents), self._slate_size):
      yield slate, tuple([documents[i] for i in slate])

  def _enumerate_state_action_indices(self, observation):
    for (slate, slate_features) in self._enumerate_slates(observation['doc']):
      state_action_pair = {
          'user': observation['user'],
          'action': slate_features
      }
      if not self._ignore_response:
        state_action_pair['response'] = observation['response']
      state_action_index = self._observation_featurizer.apply_and_flatten([
          state_action_pair,
      ])
      state_action_index = tuple(state_action_index)
      yield slate, state_action_index

  def step(self, reward, observation):
    """Records the most recent transition and returns the agent's next action.

    We store the observation of the last time step since we want to store it
    with the reward.

    Args:
      reward: The reward received from the agent's most recent action as a
        float.
      observation: A dictionary that includes the most recent observations and
        should have the following fields:
        - user: A NumPy array representing user's observed state. Assumes it is
          a concatenation of topic pull counts and topic click counts.
        - doc: A NumPy array representing observations of document features.
          Assumes it is a concatenation of one-hot encoding of topic_id and
          document quality.

    Returns:
      slate: An integer array of size _slate_size, where each element is an
        index into the list of doc_obs
    Raises:
      ValueError: if reward is not in [0, 1].
    """
    # Find max-Q action given the current state and Q-table.
    max_q_state_action = max(
        self._enumerate_state_action_indices(observation),
        key=lambda sa: self._q_value_table.get(sa[1], 0))
    max_q_next = self._q_value_table.get(max_q_state_action[1], 0)
    # Update the Q-table.
    if self._previous_state_action_index is not None:
      old_q = self._q_value_table.get(self._previous_state_action_index, 0.)
      self._q_value_table[self._previous_state_action_index] = (
          self._learning_rate * (reward + self._gamma * max_q_next) +
          (1. - self._learning_rate) * old_q)
      self._state_action_counts[
          self._previous_state_action_index] = self._state_action_counts.get(
              self._previous_state_action_index, 0) + 1
    # Pick next action.
    if not self._eval_mode:
      slate, state_action_index = self._exploration_functions[
          self._exploration_policy](
              observation)
      self._previous_state_action_index = state_action_index
    else:
      slate, state_action_index = max_q_state_action
    return slate

  def end_episode(self, reward, observation):
    self._exploration_temperature *= self._base_exploration_temperature
    self._exploration_functions = {
        'epsilon_greedy':
            lambda observation: agent_utils.epsilon_greedy_exploration(  # pylint: disable=g-long-lambda
                self._enumerate_state_action_indices(observation), self.
                _q_value_table, self._exploration_temperature),
        'min_count':
            lambda observation: agent_utils.min_count_exploration(  # pylint: disable=g-long-lambda
                self._enumerate_state_action_indices(observation),
                self._state_action_counts)
    }
    self._previous_state_action_index = None

  def bundle_and_checkpoint(self, checkpoint_dir, iteration_number):
    """Returns a self-contained bundle of the agent's state.

    Args:
      checkpoint_dir: A string for the directory where objects will be saved.
      iteration_number: An integer of iteration number to use for naming the
        checkpoint file.

    Returns:
      A dictionary containing additional Python objects to be checkpointed by
        the experiment. Each key is a string for the object name and the value
        is actual object. If the checkpoint directory does not exist, returns
        empty dictionary.
    """
    del checkpoint_dir  # Unused.
    del iteration_number  # Unused.
    bundle_dict = {'q_value_table': self._q_value_table}
    bundle_dict['sa_count'] = self._state_action_counts
    return bundle_dict

  def unbundle(self, checkpoint_dir, iteration_number, bundle_dict):
    """Restores the agent from a checkpoint.

    Args:
      checkpoint_dir: A string that represents the path to the checkpoint saved
        by tf.Save.
      iteration_number: An integer that represents the checkpoint version and is
        used when restoring replay buffer.
      bundle_dict: A dict containing additional Python objects owned by the
        agent. Each key is an object name and the value is the actual object.

    Returns:
      bool, True if unbundling was successful.
    """
    del checkpoint_dir  # Unused.
    del iteration_number  # Unused.
    if 'q_value_table' not in bundle_dict:
      logging.warning(
          'Could not unbundle from checkpoint files with exception.')
      return False
    self._q_value_table = bundle_dict['q_value_table']
    self._state_action_counts = bundle_dict.get('sa_count', {})
    return True



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for recsim.environment."""

import numpy as np
from recsim.environments import interest_exploration as ie
from recsim.simulator import environment
import tensorflow as tf


class EnvironmentTest(tf.test.TestCase):

  def setUp(self):
    super(EnvironmentTest, self).setUp()
    self._slate_size = 2
    self._num_candidates = 20
    user_model = ie.IEUserModel(
        self._slate_size,
        user_state_ctor=ie.IEUserState,
        response_model_ctor=ie.IEResponse)
    document_sampler = ie.IETopicDocumentSampler()
    self._environment = environment.Environment(user_model, document_sampler,
                                                self._num_candidates,
                                                self._slate_size)

  def test_environment(self):
    user_obs, documents = self._environment.reset()
    self.assertAllEqual(np.array([]), user_obs)
    self.assertAllEqual([
        str(doc)
        for doc in range(self._num_candidates, 2 * self._num_candidates)
    ], sorted(documents.keys()))
    slate = [0, 0]
    for i, doc_id in enumerate(list(documents)):
      if documents[doc_id]['cluster_id'] > 0.5:
        slate[0] = i
      else:
        slate[1] = i
      if slate[0] != 0 and slate[1] != 0:
        break
    user_obs, documents, _, done = self._environment.step(slate)
    self.assertAllEqual(np.array([]), user_obs)
    self.assertAllEqual([
        str(doc)
        for doc in range(2 * self._num_candidates, 3 * self._num_candidates)
    ], sorted(documents.keys()))
    self.assertFalse(done)


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for recsim.agents.random_agent."""

from gym import spaces
import numpy as np
from recsim import choice_model
from recsim.agents import random_agent
from recsim.environments import interest_evolution as iev
from recsim.environments import interest_exploration as ie
from recsim.simulator import environment
import tensorflow as tf


class RandomAgentTest(tf.test.TestCase):

  def setUp(self):
    super(RandomAgentTest, self).setUp()
    # The maximum length of videos in response
    iev.IEvResponse.MAX_VIDEO_LENGTH = 100.0

    # The number of features used to represent user state.
    iev.IEvUserState.NUM_FEATURES = 10

    # The number of features used to represent video.
    iev.IEvVideo.NUM_FEATURES = 10
    # The maximum length of videos
    iev.IEvVideo.MAX_VIDEO_LENGTH = 100.0

  def test_step(self):
    # Create a simple user
    slate_size = 2
    user_model = iev.IEvUserModel(
        slate_size,
        choice_model_ctor=choice_model.MultinomialLogitChoiceModel,
        response_model_ctor=iev.IEvResponse)

    # Create a candidate_set with 5 items
    num_candidates = 5
    document_sampler = iev.IEvVideoSampler()
    ievsim = environment.Environment(user_model, document_sampler,
                                     num_candidates, slate_size)

    # Create agent
    action_space = spaces.MultiDiscrete(num_candidates * np.ones((slate_size,)))
    agent = random_agent.RandomAgent(action_space, random_seed=0)

    # This agent doesn't use the previous user response
    observation, documents = ievsim.reset()
    slate = agent.step(1, dict(user=observation, doc=documents))
    self.assertAllEqual(slate, [2, 0])

  def test_slate_indices_and_length(self):
    # Initialize agent
    slate_size = 2
    num_candidates = 100
    action_space = spaces.MultiDiscrete(num_candidates * np.ones((slate_size,)))

    user_model = iev.IEvUserModel(
        slate_size,
        choice_model_ctor=choice_model.MultinomialLogitChoiceModel,
        response_model_ctor=iev.IEvResponse)
    agent = random_agent.RandomAgent(action_space, random_seed=0)

    # Create a set of documents
    document_sampler = iev.IEvVideoSampler()
    ievenv = environment.Environment(user_model, document_sampler,
                                     num_candidates, slate_size)

    # Test that slate indices in correct range and length is correct
    observation, documents = ievenv.reset()
    slate = agent.step(1, dict(user=observation, doc=documents))
    self.assertLen(slate, slate_size)
    self.assertAllInSet(slate, range(num_candidates))

  def test_bundle_and_unbundle_trivial(self):
    action_space = spaces.MultiDiscrete(np.ones((1,)))
    agent = random_agent.RandomAgent(action_space, random_seed=0)
    self.assertFalse(agent.unbundle('', 0, {}))
    self.assertEqual({
        'episode_num': 0
    }, agent.bundle_and_checkpoint('', 0))

  def test_bundle_and_unbundle(self):
    # Initialize agent
    slate_size = 1
    num_candidates = 3
    action_space = spaces.MultiDiscrete(num_candidates * np.ones((slate_size,)))

    user_model = ie.IEUserModel(
        slate_size,
        user_state_ctor=ie.IEUserState,
        response_model_ctor=ie.IEResponse)
    agent = random_agent.RandomAgent(action_space, random_seed=0)

    # Create a set of documents
    document_sampler = ie.IETopicDocumentSampler()
    documents = {}
    for i in range(num_candidates):
      video = document_sampler.sample_document()
      documents[i] = video.create_observation()

    # Test that slate indices in correct range and length is correct
    observation = dict(user=user_model.create_observation(), doc=documents)
    agent.step(1, observation)

    bundle_dict = agent.bundle_and_checkpoint('', 0)
    self.assertTrue(agent.unbundle('', 0, bundle_dict))
    self.assertEqual(bundle_dict, agent.bundle_and_checkpoint('', 0))


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Correlated interest exploration environment.

This environment models the problem of active exploration of user interests. It
is meant to illustrate popularity bias in recommender systems, where myopic
maximization of engagement leads to bias towards documents that have wider
appeal, whereas niche user interests remain unexplored.

In this setting, documents are generated from M topics (types) such that each
document belongs to exactly one topic. Furthermore, there are N types (types)
of users. Each document d has a production quality f_D(d) score drawn from a
distribution associated with its type D (e.g. more mass-appeal types tend
to have higher production values). On the other hand, each user u has an
affinity score g_U(u,d) towards each document type (drawn from a distribution
associated with the user's type). The final affinity of user u to document
d is thus g_U(u,d) + f_D(d). When faced with a slate of documents, the user
clicks on a document based on a multinomial logistic choice model with the
affinity scores as parameters.

A myopic agent will favor types with high production value, as they have
a high apriori probability of getting clicked across all user types. This leads
the agent to ignore niche interests, producing a suboptimal policy. This
scenario can be seen as a correlated arms bandit problem.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import flags
from absl import logging
import gin.tf
from gym import spaces
import numpy as np
from recsim import choice_model
from recsim import document
from recsim import user
from recsim import utils
from recsim.simulator import environment
from recsim.simulator import recsim_gym

FLAGS = flags.FLAGS


class IEUserModel(user.AbstractUserModel):
  """Class to model a user.

  The user in this scenario is completely characterized by a vector g
  of affinity scores of dimension |D| (the number of document topics). When
  presented with a slate of documents, the user scores each document as g(d) +
  f(d), where f(d) is the document's quality score, and then chooses according
  to a choice model based on these scores.

  The state space consists of a vector of affinity scores which is unique to the
  user and static but not observable.

  Args:
  slate_size: An integer representing the size of the slate.
  no_click_mass: A float indicating the mass given to a no-click option.
    Must be positive, otherwise CTR is always 1.
  choice_model_ctor: A contructor function to create user choice model.
  user_state_ctor: A constructor to create user state.
  response_model_ctor: A constructor function to create response. The
    function should take a string of doc ID as input and returns a
    IEResponse object.
  seed: an integer used as the seed in random sampling.
  """

  def __init__(self,
               slate_size,
               no_click_mass=5,
               choice_model_ctor=choice_model.MultinomialLogitChoiceModel,
               user_state_ctor=None,
               response_model_ctor=None,
               seed=0):
    if no_click_mass < 0:
      raise ValueError('no_click_mass must be positive.')

    super(IEUserModel, self).__init__(response_model_ctor, IEClusterUserSampler(
        user_ctor=user_state_ctor, seed=seed), slate_size)
    self._user_state_ctor = user_state_ctor
    if choice_model_ctor is None:
      raise Exception('A choice model needs to be specified!')
    self.choice_model = choice_model_ctor({
        'no_click_mass': no_click_mass,
    })

  @property
  def avg_user_state(self):
    """Returns the prior of user state."""
    return self._user_state_ctor(self._user_sampler.avg_affinity_given_topic())

  def is_terminal(self):
    """Returns a boolean indicating if the session is over."""
    return False

  # No state transitions.
  def update_state(self, slate_documents, responses):
    del slate_documents  # Unused
    del responses  # Unused
    return

  def simulate_response(self, documents):
    """Simulates the user's response to a slate of documents with choice model.

    Args:
      documents: a list of IEDocument objects in the slate.

    Returns:
      responses: a list of IEResponse objects, one for each document.
    """
    # List of empty responses
    responses = [self._response_model_ctor() for _ in documents]

    # Sample the response for selected slate items.

    self.choice_model.score_documents(
        self._user_state, [doc.create_observation() for doc in documents])
    selected_index = self.choice_model.choose_item()
    for i, response in enumerate(responses):
      response.quality = documents[i].quality
      response.cluster_id = documents[i].cluster_id
    if selected_index is None:
      return responses
    self._generate_response(documents[selected_index],
                            responses[selected_index])
    return responses

  def _generate_response(self, doc, response):
    """Trivial response: sets the clicked property of a clicked document.

    Args:
      doc: a IEDocument object.
      response: a IEResponse for the document.
    Updates: response, with whether the document was clicked.
    """
    response.clicked = True


class IEUserState(user.AbstractUserState):
  """Class to represent users.

  Attributes:
    topic_affinity: a nonnegative vector holds document type affinities which
      are not temporal dynamics and hidden.
  """

  def __init__(self, topic_affinity):
    """Initializes a new user."""
    self.topic_affinity = topic_affinity

  def score_document(self, doc_obs):
    """Returns user document affinity plus document quality."""
    return self.topic_affinity[doc_obs['cluster_id']] + doc_obs['quality']

  def create_observation(self):
    """User's topic_affinity is not observable."""
    return np.array([])

  @staticmethod
  def observation_space():
    return spaces.Box(shape=(0,), dtype=np.float32, low=0.0, high=np.inf)


@gin.configurable
class IEClusterUserSampler(user.AbstractUserSampler):
  """Samples users from predetermined types with type-specific parameters.

    This sampler consumes a distribution over user types and type-specific
    parameters for the user's affinity towards content types. It first samples
    a user type, then using that user type generates affinities according to
    the type-specific parameters. In this case, these are the mean and scale
    of a lognormal distribution, i.e. the affinity of user u of type U towards
    an document of type D is drawn according to
    lognormal(mean(U,D), scale(U,D)).

    Args:
      user_type_distribution: a non-negative array of dimension equal to the
        number of user types, whose entries sum to one.
      user_document_mean_affinity_matrix: a non-negative two-dimensional array
        with dimensions number of user types by number of document topics.
        Represents the mean of the affinity score of a user type to a topic.
      user_document_stddev_affinity_matrix: a non-negative two-dimensional array
        with dimensions number of user types by number of document topics.
        Represents the scale of the affinity score of a user type to a topic.
      user_ctor: constructor for a user state.
  """

  def __init__(self,
               user_type_distribution=(0.3, 0.7),
               user_document_mean_affinity_matrix=((.1, .7), (.7, .1)),
               user_document_stddev_affinity_matrix=((.1, .1), (.1, .1)),
               user_ctor=IEUserState,
               **kwargs):
    self._number_of_user_types = len(user_type_distribution)
    self._user_type_dist = user_type_distribution
    if len(user_document_mean_affinity_matrix) != len(user_type_distribution):
      raise ValueError('The dimensions of user_type_distribution and '
                       'user_document_mean_affinity_matrix do not match.')
    if len(user_document_stddev_affinity_matrix) != len(user_type_distribution):
      raise ValueError('The dimensions of user_type_distribution and '
                       'user_document_stddev_affinity_matrix do not match.')
    self._user_doc_means = user_document_mean_affinity_matrix
    self._user_doc_stddev = user_document_stddev_affinity_matrix
    logging.debug('Initialized IEClusterUserSampler')
    super(IEClusterUserSampler, self).__init__(user_ctor, **kwargs)

  def sample_user(self):
    # 1. Pick user type.
    user_type = self._rng.choice(
        self._number_of_user_types, p=self._user_type_dist)
    # 2. Sample user-document affinity given type.
    user_doc_affinity = (
        self._rng.lognormal(
            mean=self._user_doc_means[user_type],
            sigma=self._user_doc_stddev[user_type]))
    return self._user_ctor(user_doc_affinity)

  def avg_affinity_given_topic(self):
    # Returns the prior of document affinity.
    return np.matmul(self._user_type_dist, self._user_doc_means)


class IEResponse(user.AbstractResponse):
  """Class to represent a user's response to a document.

  Attributes:
    clicked: boolean indicating whether the item was clicked or not.
    quality: a float indicating the quality of the document.
    cluster_id: an integer representing the topic ID of the document.
  """

  NUM_CLUSTERS = 0

  def __init__(self,
               clicked=False,
               quality=0.0,
               cluster_id=0):
    self.clicked = clicked
    self.quality = quality
    self.cluster_id = cluster_id

  def __str__(self):
    return str(self.clicked)

  def __repr__(self):
    return self.__str__()

  def create_observation(self):
    return {'click': int(self.clicked),
            'quality': self.quality,
            'cluster_id': self.cluster_id}

  @classmethod
  def response_space(cls):
    return spaces.Dict({
        'click':
            spaces.Discrete(2),
        'quality':
            spaces.Box(
                low=0.0, high=np.inf, shape=tuple(), dtype=np.float32),
        'cluster_id':
            spaces.Discrete(cls.NUM_CLUSTERS)
    })


class IEDocument(document.AbstractDocument):
  """Class to represent an IE Document.

  Attributes:
    cluster_id: an integer representing the document cluster.
    quality: non-negative real number representing the quality of the document.
  """

  NUM_CLUSTERS = 0

  def __init__(self, doc_id, cluster_id, quality):
    self.cluster_id = cluster_id
    self.quality = quality
    # doc_id is an integer representing the unique ID of this document
    super(IEDocument, self).__init__(doc_id)

  def create_observation(self):
    return {'quality': self.quality,
            'cluster_id': self.cluster_id}

  @classmethod
  def observation_space(cls):
    return spaces.Dict({
        'quality':
            spaces.Box(
                low=0.0, high=np.inf, shape=tuple(), dtype=np.float32),
        'cluster_id':
            spaces.Discrete(cls.NUM_CLUSTERS)
    })


@gin.configurable
class IETopicDocumentSampler(document.AbstractDocumentSampler):
  """Samples documents with topic-specific quality distribution.

     Consumes a distribution over document topics and topic-specific parameters
     for generating a quality score (according to a lognormal distribution).

     Args:
       topic_distribution: a non-negative array of dimension equal to the
         number of topics, whose entries sum to one.
       topic_quality_mean: a non-negative array of dimension equal to the
         number of topics, representing the mean of the topic quality score.
       topic_quality_stddev: a non-negative array of dimension equal to the
         number of topics, representing the scale of the topic quality score.
       doc_ctor: A class/constructor for the type of videos that will be sampled
        by this sampler.
  """

  def __init__(self,
               topic_distribution=(.2, .8),
               topic_quality_mean=(.8, .2),
               topic_quality_stddev=(.1, .1),
               doc_ctor=IEDocument,
               **kwargs):
    self._number_of_topics = len(topic_distribution)
    self._topic_dist = topic_distribution
    if len(topic_quality_mean) != len(topic_distribution):
      raise ValueError('The dimensions of topic_quality_mean and '
                       'topic_distribution do not match.')
    if len(topic_quality_stddev) != len(topic_distribution):
      raise ValueError('The dimensions of topic_quality_stddev and '
                       'topic_distribution do not match.')
    self._topic_quality_mean = topic_quality_mean
    self._topic_quality_stddev = topic_quality_stddev
    super(IETopicDocumentSampler, self).__init__(doc_ctor, **kwargs)
    self._doc_count = 0

  @property
  def num_clusters(self):
    return self._number_of_topics

  def sample_document(self):
    """Samples the topic and then samples document features given the topic."""
    doc_features = {}
    doc_features['doc_id'] = self._doc_count
    self._doc_count += 1
    topic_id = self._rng.choice(self._number_of_topics, p=self._topic_dist)
    doc_quality = (
        self._rng.lognormal(
            mean=self._topic_quality_mean[topic_id],
            sigma=self._topic_quality_stddev[topic_id]))
    doc_features['cluster_id'] = topic_id
    doc_features['quality'] = doc_quality
    return self._doc_ctor(**doc_features)


def total_clicks_reward(responses):
  """Calculates the total number of clicks from a list of responses.

  Args:
     responses: A list of IEResponse objects

  Returns:
    reward: A float representing the total clicks from the responses
  """
  reward = 0.0
  for r in responses:
    reward += r.clicked
  return reward


def create_environment(env_config):
  """Creates an interest exploration environment."""

  document_sampler = IETopicDocumentSampler(seed=env_config['seed'])
  IEDocument.NUM_CLUSTERS = document_sampler.num_clusters
  IEResponse.NUM_CLUSTERS = document_sampler.num_clusters

  user_model = IEUserModel(
      env_config['slate_size'],
      user_state_ctor=IEUserState,
      response_model_ctor=IEResponse,
      seed=env_config['seed'])

  ieenv = environment.Environment(
      user_model,
      document_sampler,
      env_config['num_candidates'],
      env_config['slate_size'],
      resample_documents=env_config['resample_documents'])

  return recsim_gym.RecSimGymEnv(ieenv, total_clicks_reward,
                                 utils.aggregate_video_cluster_metrics,
                                 utils.write_video_cluster_metrics)



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes to represent and interface with documents."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
from gym import spaces
import numpy as np

# Some notes:
#
#   These represent properties of the documents. Presumably we can add more or
#   remove documents from the candidate set. But these do not include features
#   that depend on both the user and the document.
#


class CandidateSet(object):
  """Class to represent a collection of AbstractDocuments.

     The candidate set is represented as a hashmap (dictionary), with documents
     indexed by their document ID.
  """

  def __init__(self):
    """Initializes a document candidate set with 0 documents."""
    self._documents = {}

  def size(self):
    """Returns an integer, the number of documents in this candidate set."""
    return len(self._documents)

  def get_all_documents(self):
    """Returns all documents."""
    return self.get_documents(self._documents.keys())

  def get_documents(self, document_ids):
    """Gets the documents associated with the specified document IDs.

    Args:
      document_ids: an array representing indices into the candidate set.
        Indices can be integers or string-encoded integers.

    Returns:
      (documents) an ordered list of AbstractDocuments associated with the
        document ids.
    """
    return [self._documents[int(k)] for k in document_ids]

  def add_document(self, document):
    """Adds a document to the candidate set."""
    self._documents[document.doc_id()] = document

  def remove_document(self, document):
    """Removes a document from the set (to simulate a changing corpus)."""
    del self._documents[document.doc_id()]

  def create_observation(self):
    """Returns a dictionary of observable features of documents."""
    return {
        str(k): self._documents[k].create_observation()
        for k in self._documents.keys()
    }

  def observation_space(self):
    return spaces.Dict({
        str(k): self._documents[k].observation_space()
        for k in self._documents.keys()
    })


class AbstractDocumentSampler(object):  # pytype: disable=ignored-metaclass
  """Abstract class to sample documents."""

  __metaclass__ = abc.ABCMeta

  def __init__(self, doc_ctor, seed=0):
    self._doc_ctor = doc_ctor
    self._seed = seed
    self.reset_sampler()

  def reset_sampler(self):
    self._rng = np.random.RandomState(self._seed)

  @abc.abstractmethod
  def sample_document(self):
    """Samples and return an instantiation of AbstractDocument."""

  def get_doc_ctor(self):
    """Returns the constructor/class of the documents that will be sampled."""
    return self._doc_ctor

  @property
  def num_clusters(self):
    """Returns the number of document clusters. Returns 0 if not applicable."""
    return 0


class AbstractDocument(object):  # pytype: disable=ignored-metaclass
  """Abstract class to represent a document and its properties."""

  __metaclass__ = abc.ABCMeta

  # Number of features to represent the document.
  NUM_FEATURES = None

  def __init__(self, doc_id):
    self._doc_id = doc_id  # Unique identifier for the document

  def doc_id(self):
    """Returns the document ID."""
    return self._doc_id

  @abc.abstractmethod
  def create_observation(self):
    """Returns observable properties of this document as a float array."""

  @classmethod
  @abc.abstractmethod
  def observation_space(cls):
    """Gym space that defines how documents are represented."""



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for recsim.choice_model."""

import numpy as np
from recsim import choice_model
from recsim.environments import interest_evolution as evolution
import tensorflow as tf


class SoftmaxTest(tf.test.TestCase):

  def test_softmax_single_int(self):
    self.assertAllClose(choice_model.softmax([0]), [1.0])

  def test_softmax_equal_ints(self):
    self.assertAllClose(
        choice_model.softmax(np.ones(4)), np.array([0.25, 0.25, 0.25, 0.25]))

  def test_softmax_positive_floats(self):
    self.assertAllClose(
        choice_model.softmax(np.log(np.arange(1, 5))),
        np.array([0.1, 0.2, 0.3, 0.4]))

  def test_softmax_negative_floats(self):
    self.assertAllClose(
        choice_model.softmax(-1.0 * np.log(np.arange(1, 5))),
        np.array([0.48, 0.24, 0.16, 0.12]))


class MultinomialChoiceModelTest(tf.test.TestCase):

  def setUp(self):
    super(MultinomialChoiceModelTest, self).setUp()
    np.random.seed(0)
    self._user_state = evolution.IEvUserState(np.array([0.8, 0.6]))

  def test_multinomial_logit_default(self):
    mnl_model = choice_model.MultinomialLogitChoiceModel(choice_features={})
    mnl_model.score_documents(
        self._user_state, np.array([[0.8, 0.6], [0.6, 0.8]]))
    # The logits for two documents are 1 and 0.96 respectively. When computing
    # softmax logits, we subtract the largest value, which is 1 here. So the
    # score is softmax([0, -0.04]) = [0.51, 0.49]
    self.assertAlmostEqual(mnl_model._scores[0], 0.510, delta=0.001)
    self.assertAlmostEqual(mnl_model._scores[1], 0.490, delta=0.001)
    self.assertEqual(mnl_model._score_no_click, 0)

  def test_multinomial_logit_no_click_mass(self):
    choice_features = dict(no_click_mass=1.0)
    mnl_model = choice_model.MultinomialLogitChoiceModel(
        choice_features=choice_features)
    # The logits for two documents are 1 and 0.96 respectively. No click mass
    # is 1.0. When computing
    # softmax logits, we subtract the largest value, which is 1 here. So the
    # score is softmax([0, -0.04, 0]) = [0.337, 0.325, 0.338]
    mnl_model.score_documents(
        self._user_state, np.array([[0.8, 0.6], [0.6, 0.8]]))
    self.assertAlmostEqual(mnl_model._scores[0], 0.338, delta=0.001)
    self.assertAlmostEqual(mnl_model._scores[1], 0.325, delta=0.001)
    self.assertAlmostEqual(mnl_model._score_no_click, 0.338, delta=0.001)

  def test_multinomial_proportion_choice_model_default(self):
    choice_features = dict(min_normalizer=0)
    mnp_model = choice_model.MultinomialProportionalChoiceModel(
        choice_features=choice_features)
    mnp_model.score_documents(
        self._user_state, np.array([[0.8, 0.6], [0.6, 0.8]]))
    # The scores are the dot product between user features and doc features.
    self.assertAlmostEqual(mnp_model._scores[0], 1, delta=0.001)
    self.assertAlmostEqual(mnp_model._scores[1], 0.96, delta=0.001)
    self.assertEqual(mnp_model._score_no_click, 0)

  def test_multinomial_proportion_min_normalizer(self):
    choice_features = dict(min_normalizer=0.5, no_click_mass=0.5)
    mnp_model = choice_model.MultinomialProportionalChoiceModel(
        choice_features=choice_features)
    mnp_model.score_documents(
        self._user_state, np.array([[0.8, 0.6], [0.6, 0.8]]))
    # The scores the dot product user features and doc features minus the
    # min_normalizer.
    self.assertAlmostEqual(mnp_model._scores[0], 0.5, delta=0.001)
    self.assertAlmostEqual(mnp_model._scores[1], 0.46, delta=0.001)
    self.assertAlmostEqual(mnp_model._score_no_click, 0, delta=0.001)


class CascadeChoiceModelTest(tf.test.TestCase):

  def setUp(self):
    super(CascadeChoiceModelTest, self).setUp()
    np.random.seed(0)
    self._user_state = evolution.IEvUserState(np.array([1.0]))

  def test_exponential_cascade_invalid_score_scaling(self):
    with self.assertRaises(ValueError):
      choice_features = {'attention_prob': 0.8, 'score_scaling': -1.0}
      choice_model.ExponentialCascadeChoiceModel(choice_features)

  def test_exponential_cascade_invalid_attenion_prob(self):
    with self.assertRaises(ValueError):
      choice_features = {'attention_prob': 2.0}
      choice_model.ExponentialCascadeChoiceModel(choice_features)

  def test_exponential_cascade(self):
    choice_features = {'score_scaling': 0.04}
    model = choice_model.ExponentialCascadeChoiceModel(choice_features)
    model.score_documents(self._user_state, np.array([[3.0], [2.0], [1.0]]))
    self.assertEqual(model.choose_item(), 0)

  def test_exponential_cascade_with_no_click(self):
    choice_features = {'attention_prob': 1.0, 'score_scaling': 0.0}
    model = choice_model.ExponentialCascadeChoiceModel(choice_features)
    model.score_documents(self._user_state, np.array([[3.0], [2.0], [1.0]]))
    self.assertEqual(model.choose_item(), None)

  def test_proportional_cascade_invalid_attenion_prob(self):
    with self.assertRaises(ValueError):
      choice_features = {
          'attention_prob': 2.0,
          'min_normalizer': -2.0,
          'score_scaling': 0.1
      }
      choice_model.ProportionalCascadeChoiceModel(choice_features)

  def test_proportional_cascade_invalid_score_scaling(self):
    with self.assertRaises(ValueError):
      choice_features = {
          'attention_prob': 0.5,
          'min_normalizer': -2.0,
          'score_scaling': -1.0
      }
      choice_model.ProportionalCascadeChoiceModel(choice_features)

  def test_proportional_cascade(self):
    choice_features = {
        'attention_prob': 1.0,
        'min_normalizer': -4.0,
        'score_scaling': 0.07
    }
    model = choice_model.ProportionalCascadeChoiceModel(choice_features)
    model.score_documents(self._user_state,
                          np.array([[-3.0], [-3.0], [10.0], [1.0], [-4.0]]))
    self.assertEqual(model.choose_item(), 2)

  def test_proportional_cascade_with_no_click(self):
    choice_features = {
        'attention_prob': 0.5,
        'min_normalizer': -1.0,
        'score_scaling': 0.1
    }
    model = choice_model.ProportionalCascadeChoiceModel(choice_features)
    model.score_documents(self._user_state, np.array([[0.0], [0.0], [0.0]]))
    self.assertEqual(model.choose_item(), None)


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes for Bandit Algorithms."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy as np


class MABAlgorithm(object):
  """Base class for Multi-armed bandit (MAB) algorithms.

  We implement multi-armed bandit algorithms with confidence width tuning
  proposed in Hsu et al. https://arxiv.org/abs/1904.02664.

  Attributes:
    pulls: A numpy array which counts number of pulls of each arm
    reward: A numpy array which sums up reward of each arm
    optimism_scaling: A float specifying the confidence level. Default value
      (1.0) corresponds to the exploration strategy presented in the literature.
      A smaller number means less exploration and more exploitation.
    _rng: An instance of random.RandomState for random number generation
  """

  def __init__(self, num_arms, params, seed=0):
    """Initializes MABAlgorithm.

    Args:
      num_arms: Number of arms. Must be greater than one.
      params: A dictionary which includes additional parameters like
        optimism_scaling. Default is an empty dictionary.
      seed: Random seed for this object. Default is zero.
    """
    if num_arms < 2:
      raise ValueError('num_arms must be greater than one.')
    self.pulls = np.zeros(num_arms)
    self.reward = np.zeros(num_arms)
    self._rng = np.random.RandomState(seed)

    self.optimism_scaling = 1.0
    for attr, val in params.items():
      setattr(self, attr, val)

  def set_state(self, pulls, reward):
    if len(pulls) != len(self.pulls) or len(reward) != len(self.reward):
      raise ValueError('Cannot set state with a different number of arms.')
    self.pulls[:] = pulls
    self.reward[:] = reward

  def update(self, arm, reward):
    if reward < 0 or reward > 1:
      raise ValueError('reward must be in [0, 1].')
    self.pulls[arm] += 1
    self.reward[arm] += reward


class UCB1(MABAlgorithm):
  """UCB1 algorithm.

  See "Finite-time Analysis of the Multiarmed Bandit Problem" by Auer,
  Cesa-Bianchi, and Fischer.
  """

  def get_score(self, t):
    """Computes upper confidence bounds of reward / pulls at round t."""
    # Pull any arm that we haven't pulled.
    if not all(self.pulls):
      return np.where(self.pulls > 0, 0, np.Inf)
    ct = self.optimism_scaling * np.sqrt(2 * np.log(t))
    return self.reward / self.pulls + ct * np.sqrt(1 / self.pulls)

  def get_arm(self, t):
    return np.argmax(self.get_score(t))

  @staticmethod
  def print():
    return 'UCB1'


class KLUCB(MABAlgorithm):
  """Kullback-Leibler Upper Confidence Bounds (KL-UCB) algorithm.

  See "The KL-UCB algorithm for bounded stochastic bandits and beyond" by
  Garivier and Cappe.
  """

  def get_score(self, t):
    """Computes upper confidence bounds of reward / pulls at round t."""
    # Pull any arm that we haven't pulled.
    if not all(self.pulls):
      return np.where(self.pulls > 0, 0, np.Inf)
    c = self.optimism_scaling**2 * (np.log(t) +
                                    3 * np.log(np.log(t))) / self.pulls
    p = self.reward / self.pulls

    # KL-divergence d(p, q) is strictly increasing over [p, 1].
    # Use binary search to find q such that d(p, q) <= c.
    qmin = p
    qmax = np.ones(p.size)
    for _ in range(16):  # Error bounded by 2^-16.
      q = (qmax + qmin) / 2
      ndx = (np.where(p > 0, p * np.log(p / q), 0) +
             np.where(p < 1, (1 - p) * np.log((1 - p) / (1 - q)), 0)) < c
      qmin[ndx] = q[ndx]
      qmax[~ndx] = q[~ndx]

    return q

  def get_arm(self, t):
    return np.argmax(self.get_score(t))

  @staticmethod
  def print():
    return 'KL-UCB'


class ThompsonSampling(MABAlgorithm):
  """Thompson Sampling algorithm for the Bernoulli bandit.

  See "Further Optimal Regret Bounds for Thompson Sampling" by Agrawal and
    Goyal.
  """

  def update(self, arm, reward):
    if reward > 0 and reward < 1:
      reward = float(self._rng.rand() < reward)
    MABAlgorithm.update(self, arm, reward)

  def get_score(self, t):
    """Samples scores from the posterior distribution."""
    del t
    # Generate a beta distribution based on the expectation of reward.
    alpha = 1 + self.reward / self.optimism_scaling**2
    beta = 1 + (self.pulls - self.reward) / self.optimism_scaling**2
    return self._rng.beta(alpha, beta)

  def get_arm(self, t):
    return np.argmax(self.get_score(t))

  @staticmethod
  def print():
    return 'ThompsonSampling'



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A simple recommender system agent that recommends random slates."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from absl import logging

import numpy as np

from recsim import agent


class RandomAgent(agent.AbstractEpisodicRecommenderAgent):
  """An agent that recommends a random slate of documents."""

  def __init__(self, action_space, random_seed=0):
    super(RandomAgent, self).__init__(action_space)
    self._rng = np.random.RandomState(random_seed)

  def step(self, reward, observation):
    """Records the most recent transition and returns the agent's next action.

    We store the observation of the last time step since we want to store it
    with the reward.

    Args:
      reward: Unused.
      observation: A dictionary that includes the most recent observation.
        Should include 'doc' field that includes observation of all candidates.

    Returns:
      slate: An integer array of size _slate_size, where each element is an
        index into the list of doc_obs
    """
    del reward  # Unused argument.
    doc_obs = observation['doc']

    # Simulate a random slate
    doc_ids = list(range(len(doc_obs)))
    self._rng.shuffle(doc_ids)
    slate = doc_ids[:self._slate_size]
    logging.debug('Recommended slate: %s', slate)
    return slate



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for recsim.agents.bandits.algorithm."""

import numpy as np
from recsim.agents.bandits import algorithms
import tensorflow as tf


class UCB1Test(tf.test.TestCase):

  def setUp(self):
    super(UCB1Test, self).setUp()
    self._alg = algorithms.UCB1(2, {})
    self._alg.set_state(np.array([10, 10]), np.array([10, 0]))

  def test_get_score(self):
    ucb = self._alg.get_score(20)
    self.assertAlmostEqual(1.7740455, ucb[0])
    self.assertAlmostEqual(0.7740455, ucb[1])

  def test_get_arm(self):
    # Arm 0 is clearly the best.
    self.assertEqual(0, self._alg.get_arm(20))

  def test_get_arm_initial_stage(self):
    alg = algorithms.UCB1(3, {})
    alg.update(0, 1)
    alg.update(2, 1)
    # Pull Arm 1 which has not been pulled.
    self.assertEqual(1, alg.get_arm(1))


class KLUCBTest(tf.test.TestCase):

  def setUp(self):
    super(KLUCBTest, self).setUp()
    self._alg = algorithms.KLUCB(2, {})
    self._alg.set_state(np.array([10, 10]), np.array([10, 0]))

  def test_get_score(self):
    ucb = self._alg.get_score(20)
    self.assertAlmostEqual(1, ucb[0])
    self.assertAlmostEqual(0.5 - 2**-16, ucb[1])

  def test_get_arm(self):
    # Arm 0 is clearly the best.
    self.assertEqual(0, self._alg.get_arm(20))

  def test_get_arm_initial_stage(self):
    alg = algorithms.KLUCB(3, {})
    alg.update(0, 1)
    alg.update(2, 1)
    # Pull Arm 1 which has not been pulled.
    self.assertEqual(1, alg.get_arm(1))


class ThompsonSamplingTest(tf.test.TestCase):

  def setUp(self):
    super(ThompsonSamplingTest, self).setUp()
    self._alg = algorithms.ThompsonSampling(2, {})
    self._alg.set_state(np.array([10, 10]), np.array([10, 0]))

  def test_update(self):
    self._alg.update(0, 0.9)
    self._alg.update(1, 0)
    self.assertAllEqual(np.array([11, 11]), self._alg.pulls)
    self.assertAllEqual(np.array([11, 0]), self._alg.reward)

  def test_get_score(self):
    mu = self._alg.get_score(20)
    self.assertAlmostEqual(0.9570183, mu[0])
    self.assertAlmostEqual(0.0438080, mu[1])

  def test_get_arm(self):
    # Arm 0 is clearly the best.
    self.assertEqual(0, self._alg.get_arm(20))


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for recsim.agents.cluster_bandit_agent."""

from gym import spaces
import numpy as np
from recsim.agents import cluster_bandit_agent
from recsim.environments import interest_exploration as ie
import tensorflow as tf


class ClusterBanditAgentTest(tf.test.TestCase):

  def dummy_observation_space(self):
    single_response_space = spaces.Dict({
        'cluster_id': spaces.Discrete(2),
        'click': spaces.Discrete(2)
    })
    doc_space = spaces.Dict(
        {0: spaces.Dict({'cluster_id': spaces.Discrete(2)})})
    user_space = spaces.Dict({
        'sufficient_statistics':
            spaces.Dict({
                'impression_count':
                    spaces.Box(np.array([0] * 2), np.array([np.inf] * 2)),
                'click_count':
                    spaces.Box(np.array([0] * 2), np.array([np.inf] * 2))
            })
    })
    return spaces.Dict({
        'user': user_space,
        'doc': doc_space,
        'response': spaces.Tuple([
            single_response_space,
        ])
    })

  def doc_user_to_sufficient_stats(self, docs, observation):
    sufficient_stats_observation = {'user': {'sufficient_statistics': {}}}
    sufficient_stats_observation['user']['sufficient_statistics'][
        'impression_count'] = observation[:2]
    sufficient_stats_observation['user']['sufficient_statistics'][
        'click_count'] = observation[2:]
    sufficient_stats_observation['doc'] = docs
    return sufficient_stats_observation

  def test_step_with_bigger_slate(self):
    # Initialize agent.
    slate_size = 5
    num_candidates = 5
    action_space = spaces.MultiDiscrete(num_candidates * np.ones((slate_size,)))
    agent = cluster_bandit_agent.ClusterBanditAgent(
        self.dummy_observation_space(), action_space)

    # Create a set of documents
    document_sampler = ie.IETopicDocumentSampler(seed=1)
    documents = {}
    for i in range(num_candidates):
      video = document_sampler.sample_document()
      documents[i] = video.create_observation()

    # Past observation shows Topic 1 is better.
    user_obs = np.array([1, 1, 0, 1])
    sufficient_stats_observation = self.doc_user_to_sufficient_stats(
        documents, user_obs)
    slate = agent.step(0, sufficient_stats_observation)
    # Documents in Topic 0 sorted by quality: 1, 2.
    # Documents in Topic 1 sorted by quality: 0, 4, 3.
    self.assertAllEqual(slate, [0, 4, 3, 1, 2])

  def test_bundle_and_unbundle_trivial(self):
    action_space = spaces.MultiDiscrete(2 * np.ones((2,)))
    agent = cluster_bandit_agent.ClusterBanditAgent(
        self.dummy_observation_space(), action_space)
    self.assertFalse(agent.unbundle('', 0, {}))
    self.assertEqual(
        {
            'base_agent_bundle_0': {
                'episode_num': 0
            },
            'base_agent_bundle_1': {
                'episode_num': 0
            }
        }, agent.bundle_and_checkpoint('', 0))

  def test_bundle_and_unbundle(self):
    # Initialize agent
    slate_size = 2
    num_candidates = 5
    action_space = spaces.MultiDiscrete(num_candidates * np.ones((slate_size,)))

    agent = cluster_bandit_agent.ClusterBanditAgent(
        self.dummy_observation_space(), action_space)

    # Create a set of documents
    document_sampler = ie.IETopicDocumentSampler()
    documents = {}
    for i in range(num_candidates):
      video = document_sampler.sample_document()
      documents[i] = video.create_observation()

    # Test that slate indices in correct range and length is correct
    sufficient_stats_observation = self.doc_user_to_sufficient_stats(
        documents, np.array([0, 0, 0, 0]))

    agent.step(1, sufficient_stats_observation)

    bundle_dict = agent.bundle_and_checkpoint('', 0)
    self.assertTrue(agent.unbundle('', 0, bundle_dict))
    self.assertEqual(bundle_dict, agent.bundle_and_checkpoint('', 0))


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Abstract interface for recommender system agents."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
from absl import logging


class AbstractRecommenderAgent(object):  # pytype: disable=ignored-metaclass
  """Abstract class to model a recommender system agent."""

  __metaclass__ = abc.ABCMeta

  def __init__(self, action_space):
    """Initializes AbstractRecommenderAgent.

    Args:
      action_space: A gym.spaces object that specifies the format of actions.
    """
    self._slate_size = action_space.nvec.shape[0]

  @abc.abstractmethod
  def step(self, reward, observation):
    """Records the most recent transition and returns the agent's next action.

    We store the observation of the last time step since we want to store it
    with the reward.

    Args:
      reward: The reward received from the agent's most recent action as a
        float.
      observation: A dictionary that includes the most recent observations.

    Returns:
      slate: An integer array of size _slate_size, where each element is an
        index into the list of doc_obs
    """

  @abc.abstractmethod
  def bundle_and_checkpoint(self, checkpoint_dir, iteration_number):
    """Returns a self-contained bundle of the agent's state.

    This is used for checkpointing. It will return a dictionary containing all
    non-TensorFlow objects (to be saved into a file by the caller), and it saves
    all TensorFlow objects into a checkpoint file.

    Args:
      checkpoint_dir: A string for the directory where objects will be saved.
      iteration_number: An integer of iteration number to use for naming the
        checkpoint file.

    Returns:
      A dictionary containing additional Python objects to be checkpointed by
        the experiment. Each key is a string for the object name and the value
        is actual object. If the checkpoint directory does not exist, returns
        empty dictionary.
    """

  @abc.abstractmethod
  def unbundle(self, checkpoint_dir, iteration_number, bundle_dict):
    """Restores the agent from a checkpoint.

    Restores the agent's Python objects to those specified in bundle_dict,
    and restores the TensorFlow objects to those specified in the
    checkpoint_dir. If the checkpoint_dir does not exist, will not reset the
    agent's state.

    Args:
      checkpoint_dir: A string that represents the path to the checkpoint saved
        by tf.Save.
      iteration_number: An integer that represents the checkpoint version and is
        used when restoring replay buffer.
      bundle_dict: A dict containing additional Python objects owned by the
        agent. Each key is an object name and the value is the actual object.

    Returns:
      bool, True if unbundling was successful.
    """


class AbstractEpisodicRecommenderAgent(AbstractRecommenderAgent):
  """Abstract class for recommender systems that solves episodic tasks."""

  def __init__(self, action_space, summary_writer=None):
    """Initializes AbstractEpisodicRecommenderAgent.

    Args:
      action_space: A gym.spaces object that specifies the format of actions.
      summary_writer: A Tensorflow summary writer to pass to the agent
        for in-agent training statistics in Tensorboard.
    """
    super(AbstractEpisodicRecommenderAgent, self).__init__(action_space)
    self._episode_num = 0
    self._summary_writer = summary_writer

  def begin_episode(self, observation=None):
    """Returns the agent's first action for this episode.

    Args:
      observation: numpy array, the environment's initial observation.

    Returns:
      slate: An integer array of size _slate_size, where each element is an
        index into the list of doc_obs
    """
    self._episode_num += 1
    return self.step(0, observation)

  def end_episode(self, reward, observation=None):
    """Signals the end of the episode to the agent.

    Args:
      reward: An float that is the last reward from the environment.
      observation: numpy array that represents the last observation of the
        episode.
    """
    pass

  def bundle_and_checkpoint(self, checkpoint_dir, iteration_number):
    """Returns a self-contained bundle of the agent's state.

    Args:
      checkpoint_dir: A string that represents the path to the checkpoint and is
        used when we save TensorFlow objects by tf.Save.
      iteration_number: An integer that represents the checkpoint version and is
        used when restoring replay buffer.

    Returns:
      A dictionary containing additional Python objects to be checkpointed by
        the experiment. Each key is a string for the object name and the value
        is actual object. If the checkpoint directory does not exist, returns
        empty dictionary.
    """
    del checkpoint_dir  # Unused.
    del iteration_number  # Unused.
    bundle_dict = {}
    bundle_dict['episode_num'] = self._episode_num
    return bundle_dict

  def unbundle(self, checkpoint_dir, iteration_number, bundle_dict):
    """Restores the agent from a checkpoint.

    Args:
      checkpoint_dir: A string that represents the path to the checkpoint and is
        used when we save TensorFlow objects by tf.Save.
      iteration_number: An integer that represents the checkpoint version and is
        used when restoring replay buffer.
      bundle_dict: A dict containing additional Python objects owned by the
        agent. Each key is an object name and the value is the actual object.

    Returns:
      bool, True if unbundling was successful.
    """
    del checkpoint_dir  # Unused.
    del iteration_number  # Unused.
    if 'episode_num' not in bundle_dict:
      logging.warning(
          'Could not unbundle from checkpoint files with exception.')
      return False
    self._episode_num = bundle_dict['episode_num']
    return True


class AbstractHierarchicalAgentLayer(AbstractRecommenderAgent):
  """Parent class for stackable agent layers."""

  def __init__(self, action_space, *base_agent_ctors):
    super(AbstractHierarchicalAgentLayer, self).__init__(action_space)
    self._base_agent_ctors = base_agent_ctors
    self._base_agents = None

  def _preprocess_reward_observation(self, reward, observation):
    r"""Modifies the reward and observation before passing to base agent.

    This function is used to modify the observation and reward before
    propagating it downward to the base agent. For example, it can
    inject additional features like sufficient statistics by inserting fields
    to observation[\'user\'], or, to implement regularization schemes by
    subtracting penalties from the reward.

    Args:
      reward: float number.
      observation: gym space in recsim format.

    Returns:
      reward: float number.
      observation: gym space in recsim format.
    """
    return reward, observation

  @abc.abstractmethod
  def _postprocess_actions(self, action_list):
    r"""Aggregates (possibly abstract) base agent actions into a valid slate."""

  def begin_episode(self, observation=None):
    if observation is not None:
      _, observation = self._preprocess_reward_observation(0, observation)
    action_list = [
        base_agent.begin_episode(observation=observation)
        for base_agent in self._base_agents
    ]
    return self._postprocess_actions(action_list)

  def end_episode(self, reward, observation):
    reward, observation = self._preprocess_reward_observation(
        reward, observation)
    action_list = [
        base_agent.end_episode(reward, observation=observation)
        for base_agent in self._base_agents
    ]
    return self._postprocess_actions(action_list)

  def bundle_and_checkpoint(self, checkpoint_dir, iteration_number):
    """Returns a self-contained bundle of the agent's state.

    Args:
      checkpoint_dir: A string for the directory where objects will be saved.
      iteration_number: An integer of iteration number to use for naming the
        checkpoint file.

    Returns:
      A dictionary containing additional Python objects to be checkpointed by
        the experiment. Each key is a string for the object name and the value
        is actual object. If the checkpoint directory does not exist, returns
        empty dictionary.
    """
    bundle_dict = {}
    for i, base_agent in enumerate(self._base_agents):
      base_bundle_dict = base_agent.bundle_and_checkpoint(
          checkpoint_dir, iteration_number)
      bundle_dict['base_agent_bundle_{}'.format(i)] = base_bundle_dict
    return bundle_dict

  def unbundle(self, checkpoint_dir, iteration_number, bundle_dict):
    """Restores the agent from a checkpoint.

    Args:
      checkpoint_dir: A string that represents the path to the checkpoint saved
        by tf.Save.
      iteration_number: An integer that represents the checkpoint version and is
        used when restoring replay buffer.
      bundle_dict: A dict containing additional Python objects owned by the
        agent. Each key is an object name and the value is the actual object.

    Returns:
      bool, True if unbundling was successful.
    """
    success = True
    for i, base_agent in enumerate(self._base_agents):
      if 'base_agent_bundle_{}'.format(i) not in bundle_dict:
        logging.warning('Base agent bundle not found in bundle.')
        return False
      success &= base_agent.unbundle(
          checkpoint_dir, iteration_number,
          bundle_dict['base_agent_bundle_{}'.format(i)])
    return success



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Recsim-specific Dopamine DQN agent and related utilities."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dopamine.agents.dqn import dqn_agent
from dopamine.replay_memory import circular_replay_buffer
import gin.tf
from gym import spaces
import numpy as np
import tensorflow as tf

DQNNetworkType = collections.namedtuple('dqn_network', ['q_values'])


class ResponseAdapter(object):
  """Custom flattening of responses to accommodate dopamine replay buffer."""

  def __init__(self, input_response_space):
    """Init function for ResponseAdapter.

    Args:
      input_response_space: this is assumed to be an instance of
        gym.spaces.Tuple; each element of the tuple is has to be an instance
        of gym.spaces.Dict consisting of feature_name: 0-d gym.spaces.Box
          (single float) key-value pairs.
    """
    self._input_response_space = input_response_space
    self._single_response_space = input_response_space.spaces[0]
    self._response_names = list(self._single_response_space.spaces.keys())
    self._response_shape = (len(input_response_space.spaces),
                            len(self._response_names))
    self._response_dtype = np.float32

  @property
  def response_names(self):
    return self._response_names

  @property
  def response_shape(self):
    return self._response_shape

  @property
  def response_dtype(self):
    return self._response_dtype

  def encode(self, responses):
    response_tensor = np.zeros(self._response_shape, dtype=self._response_dtype)
    for i, response in enumerate(responses):
      # Note: the order of dictionary keys in the self._input_response_space is
      # not necessarily the same as the order of the keys in the observation
      # dictionary itself. To guarrantee the position of elements are consistent
      # with the order in self._response_names, we iterate over keys explicitly.
      for j, key in enumerate(self.response_names):
        response_tensor[i, j] = response[key]
    return response_tensor


@gin.configurable
class ObservationAdapter(object):
  """An adapter to convert between user/doc observation and images."""

  def __init__(self, input_observation_space, stack_size=1):
    self._input_observation_space = input_observation_space
    user_space = input_observation_space.spaces['user']
    doc_space = input_observation_space.spaces['doc']
    self._num_candidates = len(doc_space.spaces)

    doc_space_shape = spaces.flatdim(list(doc_space.spaces.values())[0])
    # Use the longer of user_space and doc_space as the shape of each row.
    obs_shape = (np.max([spaces.flatdim(user_space), doc_space_shape]),)
    self._observation_shape = (self._num_candidates + 1,) + obs_shape
    self._observation_dtype = user_space.dtype
    self._stack_size = stack_size

  # Pads an array with zeros to make its shape identical to  _observation_shape.
  def _pad_with_zeros(self, array):
    width = self._observation_shape[1] - len(array)
    return np.pad(array, (0, width), mode='constant')

  @property
  def output_observation_space(self):
    """The output observation space of the adapter."""
    user_space = self._input_observation_space.spaces['user']
    doc_space = self._input_observation_space.spaces['doc']
    user_dim = spaces.flatdim(user_space)
    low = np.concatenate(
        [self._pad_with_zeros(np.ones(user_dim) * -np.inf).reshape(1, -1)] + [
            self._pad_with_zeros(np.ones(spaces.flatdim(d)) *
                                 -np.inf).reshape(1, -1)
            for d in doc_space.spaces.values()
        ])
    high = np.concatenate(
        [self._pad_with_zeros(np.ones(user_dim) * np.inf).reshape(1, -1)] + [
            self._pad_with_zeros(np.ones(spaces.flatdim(d)) *
                                 np.inf).reshape(1, -1)
            for d in doc_space.spaces.values()
        ])
    return spaces.Box(low=low, high=high, dtype=np.float32)

  def encode(self, observation):
    """Encode user observation and document observations to an image."""
    # It converts the observation from the simulator to a numpy array to be
    # consumed by DQN agent, which assume the input is a "image".
    # The first row is user's observation. The remaining rows are documents'
    # observation, one row for each document.
    image = np.zeros(
        self._observation_shape + (self._stack_size,),
        dtype=self._observation_dtype)
    image[0, :, 0] = self._pad_with_zeros(
        spaces.flatten(self._input_observation_space.spaces['user'],
                       observation['user']))
    doc_space = zip(self._input_observation_space.spaces['doc'].spaces.values(),
                    observation['doc'].values())
    image[1:, :, 0] = np.array([
        self._pad_with_zeros(spaces.flatten(doc_space, d))
        for doc_space, d in doc_space
    ])

    return image


# The following functions creates the DQN network for RecSim.
def recsim_dqn_network(user, doc, scope):
  inputs = tf.concat([user, doc], axis=1)
  with tf.compat.v1.variable_scope(scope, reuse=tf.compat.v1.AUTO_REUSE):
    hidden = tf.compat.v1.layers.dense(inputs, 256, activation=tf.nn.relu)
    hidden = tf.compat.v1.layers.dense(hidden, 32, activation=tf.nn.relu)
    q_value = tf.compat.v1.layers.dense(hidden, 1, name='output')
  return q_value


class DQNAgentRecSim(dqn_agent.DQNAgent):
  """RecSim-specific Dopamine DQN agent that converts the observation space."""

  def __init__(self, sess, observation_space, num_actions, stack_size,
               optimizer_name, eval_mode, **kwargs):
    if stack_size != 1:
      raise ValueError(
          'Invalid stack_size: %s. Only stack_size=1 is supported for now.' %
          stack_size)

    self._env_observation_space = observation_space
    # In our case, the observation is a data structure that stores observation
    # of the user and candidate documents. We uses an observation adapter to
    # convert it to an "image", which is required by dopamine DQNAgent.
    self._obs_adapter = ObservationAdapter(self._env_observation_space)

    if optimizer_name == 'adam':
      optimizer = tf.compat.v1.train.AdamOptimizer()
    elif optimizer_name == 'sgd':
      optimizer = tf.compat.v1.train.GradientDescentOptimizer(
          learning_rate=1e-3)
    else:
      optimizer = tf.compat.v1.train.RMSPropOptimizer(
          learning_rate=0.00025,
          decay=0.95,
          momentum=0.0,
          epsilon=0.00001,
          centered=True)

    dqn_agent.DQNAgent.__init__(
        self,
        sess,
        num_actions,
        observation_shape=self._obs_adapter.output_observation_space.shape,
        observation_dtype=self._obs_adapter.output_observation_space.dtype,
        stack_size=stack_size,
        network=recsim_dqn_network,
        optimizer=optimizer,
        eval_mode=eval_mode,
        **kwargs)

  def _validate_states(self, states):
    shape = states.get_shape()
    if len(shape) != 4 or shape[1] != self._num_candidates + 1:
      raise ValueError('Invalid states shape: %s. '
                       'Expecting [batch_size, %s, num_features, stack_size].' %
                       (shape, self._num_candidates + 1))


def wrapped_replay_buffer(**kwargs):
  return circular_replay_buffer.WrappedReplayBuffer(**kwargs)
